import os
import yadageschemas
import logging

from . import workflow_loader as workflow_loader

from .handlers.utils import handler_decorator
from .state_providers import state_provider_from_string
from .wflowstate import load_model_fromstring
from .controllers import setup_controller
from .wflow import YadageWorkflow

log = logging.getLogger(__name__)

handlers, creator = handler_decorator()


log = logging.getLogger(__name__)


@creator("local")
def local_workflows(
    metadir,
    workflow=None,
    initdata=None,
    toplevel=os.getcwd(),
    dataarg=None,
    dataopts=None,
    wflowopts=None,
    workflow_json=None,
    modelsetup="inmem",
    modelopts=None,
    controller="frommodel",
    ctrlopts=None,
    validate=True,
    schemadir=yadageschemas.schemadir,
):
    """
    load workflow from spec and initialize it

    :param workflow: the workflow spec source
    :param toplevel: base URI against which to resolve JSON references in the spec
    :param initdata: initialization data for workflow

    prepares initial workflow object and returns controller
    """
    rootprovider = state_provider_from_string(dataarg, dataopts)

    if not workflow_json and not workflow:
        raise RuntimeError(
            "need to provide either direct workflow spec or source to load from"
        )

    if workflow_json:
        if validate:
            workflow_loader.validate(workflow_json)
    else:
        workflow_json = workflow_loader.workflow(
            workflow, toplevel=toplevel, schemadir=schemadir, validate=validate
        )

    # with open('{}/yadage_template.json'.format(metadir), 'w') as f:
    #     json.dump(workflow_json, f)
    workflowobj = YadageWorkflow.createFromJSON(workflow_json, rootprovider)
    if initdata:
        wflowopts = wflowopts or {}
        discover = wflowopts.setdefault("discover", True)
        relative = wflowopts.setdefault("relative", True)
        log.info(
            "initializing workflow with initdata: %s discover: %s relative: %s",
            initdata,
            discover,
            relative,
        )
        workflowobj.view().init(
            initdata, rootprovider, discover=discover, relative=relative
        )
    else:
        log.info("no initialization data")

    model = load_model_fromstring(modelsetup, modelopts, workflowobj)
    return setup_controller(model=model, controller=controller, ctrlopts=ctrlopts)



import logging

import jsonpath_rw
from jsonpointer import JsonPointer

from .stages import JsonStage, OffsetStage
from .utils import get_obj_id, init_stage_spec
from .wflownode import YadageNode

log = logging.getLogger(__name__)


class WorkflowView(object):
    """
    Provides a 'view' of the overall workflow object that corresponds
    to a particular level of nesting. That is, this presents the scope
    within which extension rules operate (i.e. add steps, reference
    other steps, etc)
    """

    def __init__(self, workflowobj, offset=""):
        self.wflow = workflowobj
        self.dag = workflowobj.dag
        self.rules = workflowobj.rules
        self.applied_rules = workflowobj.applied_rules
        self.offset = offset
        self.steps = JsonPointer(self.offset).resolve(workflowobj.stepsbystage)
        self.bookkeeper = JsonPointer(self.offset).resolve(workflowobj.bookkeeping)
        self.values = JsonPointer(self.offset).resolve(workflowobj.values)

    def view(self, offset):
        """
        return a view with an additional offset to this view's offset
        """
        return self.wflow.view(self._makeoffset(offset))

    def query(self, query, collection):
        """
        :return
        """
        matches = jsonpath_rw.parse(query).find(collection)
        return matches

    def getScopes(self, query):
        matches = self.query(query, self.steps)
        scopes = []
        for match in matches:
            value = match.value
            # step endpoint case
            if isinstance(value, dict) and "_offset" in value:
                scopes.append(value["_offset"])
            # stage endpoint case
            elif isinstance(value, list):
                for item in value:
                    if "_offset" in item:
                        scopes.append(item["_offset"])

        return scopes

    def getSteps(self, query):
        """
        returns steps related to the JSONPath query.
        if a query points to a stage (say 'stagename'):
        will return all toplevel steps.. no recursion into subworkflows
        if a query points to steps (e.g. 'stagename[*]')
        will return a steps directly

        """
        nodeids = []

        matches = self.query(query, self.steps)
        for match in matches:
            value = match.value
            # step endpoint case
            if isinstance(value, dict) and "_nodeid" in value:
                nodeids.append(value["_nodeid"])
            # stage endpoint case
            elif isinstance(value, list):
                for item in value:
                    if "_nodeid" in item:
                        nodeids.append(item["_nodeid"])

        result = [self.dag.getNode(nodeid) for nodeid in nodeids]
        return result

    def _makeoffset(self, offset):
        """
        prepare a full offset based on this views' offset and a relative offset

        :param offset: the relative offset
        """
        thisoffset = JsonPointer(offset)
        if self.offset:
            fulloffset = JsonPointer.from_parts(
                JsonPointer(self.offset).parts + thisoffset.parts
            ).path
        else:
            fulloffset = thisoffset.path
        return fulloffset

    def getRule(self, name=None, offset="", identifier=None):
        """retrieve a rule by offset or name or identifier"""
        fulloffset = self._makeoffset(offset)
        for x in self.rules + self.applied_rules:
            if x.identifier == identifier or (
                x.offset == fulloffset and x.rule.name == name
            ):
                return x
        return None

    def init(
        self,
        initdata,
        init_provider=None,
        used_inputs=None,
        name="init",
        discover=False,
        relative=True,
    ):
        """
        initialize this scope by adding an initialization stage.

        :param inidata: in
        itialization JSON data
        """
        spec = init_stage_spec(
            initdata, discover, used_inputs or [], name, relative=relative
        )
        self.addRule(JsonStage(spec, init_provider), self.offset)

    def addRule(self, rule, offset="", identifier=None):
        """
        add a DAG extension rule, possibly with a scope offset
        """
        thisoffset = JsonPointer(offset)
        if offset != "":
            createIndexData(thisoffset.path, self.steps, self.values)
        createOffsetMeta(thisoffset.path, self.bookkeeper)

        offsetstage = OffsetStage(rule, self._makeoffset(offset), identifier=identifier)
        self.rules += [offsetstage]
        thisoffset.resolve(self.bookkeeper)["_meta"]["stages"] += [
            offsetstage.identifier
        ]
        return offsetstage.identifier

    def addValue(self, key, value):
        v = self.values.setdefault("_values", {})
        if key in v:
            raise RuntimeError("cannot overwrite value")
        v[key] = value

    def getValue(self, key):
        return self.values.setdefault("_values", {}).get(key)

    def addStep(self, task, stage, depends_on=None):
        """
        adds a node to the DAG connecting it to the passed depending nodes
        while tracking that it was added by the specified stage

        :param task: the task object for the step
        :param stage: the stage name
        :param depends_on: dependencies of this step
        """

        self.steps.setdefault(stage, [])
        node = YadageNode(task.metadata["name"], task, identifier=get_obj_id(task))
        node.task.metadata["wflow_node_id"] = node.identifier
        node.task.metadata["wflow_offset"] = self.offset
        node.task.metadata["wflow_stage"] = stage
        node.task.metadata["wflow_stage_node_idx"] = len(self.steps[stage])
        node.task.metadata["wflow_hints"] = {"is_purepub": task.pubOnlyTask()}

        self.dag.addNode(node, depends_on=depends_on)
        self.steps[stage].append({"_nodeid": node.identifier})
        self.bookkeeper["_meta"]["steps"] += [node.identifier]
        log.info("added %s", node)
        return node

    def addWorkflow(self, rules, stage=None):
        """
        add a (sub-)workflow (i.e. list of stages) to the overall workflow
        """
        offset = ""
        if stage is not None:
            # make sure storage for the 'authoring' stage is present and
            # register the workflow as part of that 'author'
            # needed e.g. for predicate handlers trying to determing if
            # the author stage is done
            nextindex = len(self.steps.get(stage, []))
            offset = JsonPointer.from_parts([stage, nextindex]).path

            self.steps.setdefault(stage, []).append({})
            self.values.setdefault(stage, []).append({})

        for rule in rules:
            self.addRule(rule, offset)


def createIndexData(offset, stepindex, valueindex):
    pointer = JsonPointer(offset)
    pointer.resolve(stepindex)["_offset"] = offset
    pointer.set(valueindex, {})


def createOffsetMeta(offset, bookkeeping):
    """
    sets up a location to track rule and step ids for a given scope offset
    """
    pointer = JsonPointer(offset)
    view = bookkeeping
    for x in pointer.parts:
        view = view.setdefault(x, {})
    pointer.resolve(bookkeeping).setdefault("_meta", {"stages": [], "steps": []})



#!/usr/bin/env python
import logging
import os
import sys
import click
import yaml

import yadage.steering_api as steering_api
import yadage.utils as utils
import yadageschemas
from packtivity.plugins import enable_plugins

from .logutils import LOGFORMAT

log = logging.getLogger(__name__)


RC_FAILED = 1
RC_SUCCEEDED = 0
DEFAULT_BACKEND = "multiproc:auto"


def from_file(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    data = {}
    for v in value:
        data.update(**yaml.safe_load(v))
    verbosity = data.pop("verbosity", "INFO")
    logging.basicConfig(level=getattr(logging, verbosity), format=LOGFORMAT)
    enable_plugins(data.pop("plugins", []))

    data["backend"] = utils.setupbackend_fromstring(
        data.pop("backend", DEFAULT_BACKEND), data.pop("backendopts", {})
    )

    rc = RC_FAILED
    try:
        steering_api.run_workflow(**data)
        rc = RC_SUCCEEDED
    except:
        log.exception("workflow failed")
    if rc != RC_SUCCEEDED:
        exc = click.exceptions.ClickException(click.style("Workflow failed", fg="red"))
        exc.exit_code = rc
        raise exc

    ctx.exit()


@click.command()
@click.argument("dataarg", default="workdir")
@click.argument("workflow", default="workflow.yml")
@click.argument("initfiles", nargs=-1)
@click.option(
    "-b", "--backend", default=DEFAULT_BACKEND, help="packtivity backend string"
)
@click.option("-c", "--cache", default="")
@click.option(
    "-d",
    "--dataopt",
    multiple=True,
    default=None,
    help="options for the workflow data state",
)
@click.option(
    "-e",
    "--schemadir",
    default=yadageschemas.schemadir,
    help="schema directory for workflow validation",
)
@click.option(
    "-f",
    "--from-file",
    expose_value=False,
    multiple=True,
    type=click.File("rb"),
    help="read entire configuration from file, no other flags settings are read.",
    callback=from_file,
    is_eager=True,
)
@click.option("-g", "--strategy", help="set execution stragegy")
@click.option(
    "-i", "--loginterval", default=30, help="adage tracking interval in seconds"
)
@click.option(
    "-k",
    "--backendopt",
    multiple=True,
    default=None,
    help="options for the workflow data state",
)
@click.option(
    "-l",
    "--modelopt",
    multiple=True,
    default=None,
    help="options for the workflow state models",
)
@click.option(
    "-m", "--metadir", default=None, help="directory to store workflow metadata"
)
@click.option(
    "-o",
    "--ctrlopt",
    multiple=True,
    default=None,
    help="options for the workflow controller",
)
@click.option(
    "-p",
    "--parameter",
    multiple=True,
    help="<parameter name>=<yaml string> input parameter specifcations ",
)
@click.option("-r", "--controller", default="frommodel", help="controller")
@click.option("-s", "--modelsetup", default="inmem", help="wflow state model")
@click.option(
    "-t",
    "--toplevel",
    default=os.getcwd(),
    help="toplevel uri to be used to resolve workflow name and references from",
)
@click.option(
    "-u",
    "--updateinterval",
    default=0.02,
    help="adage graph inspection interval in seconds",
)
@click.option("-v", "--verbosity", default="INFO", help="logging verbosity")
@click.option(
    "-w",
    "--wflowopt",
    multiple=True,
    default=None,
    help="options for the workflow initialization",
)
@click.option(
    "-y", "--strategyopt", help="strategy option", multiple=True, default=None
)
@click.option("--accept-metadir/--no-accept-metadir", default=False)
@click.option("--plugins", default=None)
@click.option(
    "--validate/--no-validate",
    default=True,
    help="en-/disable workflow spec validation",
)
@click.option(
    "--visualize/--no-visualize", default=False, help="visualize workflow graph"
)
def main(
    dataarg,
    workflow,
    initfiles,
    controller,
    ctrlopt,
    toplevel,
    verbosity,
    loginterval,
    updateinterval,
    schemadir,
    backend,
    dataopt,
    wflowopt,
    backendopt,
    strategy,
    strategyopt,
    modelsetup,
    modelopt,
    metadir,
    parameter,
    validate,
    visualize,
    cache,
    plugins,
    accept_metadir,
):

    if os.path.exists("input.yml") and not initfiles:
        initfiles = ("input.yml",)
    logging.basicConfig(level=getattr(logging, verbosity), format=LOGFORMAT)
    from packtivity.plugins import enable_plugins

    if plugins:
        enable_plugins(plugins.split(","))

    initdata = utils.getinit_data(initfiles, parameter)
    dataopts = utils.options_from_eqdelimstring(dataopt)
    backendopts = utils.options_from_eqdelimstring(backendopt)
    modelopts = utils.options_from_eqdelimstring(modelopt)
    ctrlopts = utils.options_from_eqdelimstring(ctrlopt)
    wflowopts = utils.options_from_eqdelimstring(wflowopt)
    strategyopts = utils.options_from_eqdelimstring(strategyopt)

    backend = utils.setupbackend_fromstring(backend, backendopts)
    rc = RC_FAILED
    try:
        steering_api.run_workflow(
            workflow=workflow,
            toplevel=toplevel,
            validate=validate,
            schemadir=schemadir,
            initdata=initdata,
            wflowopts=wflowopts,
            controller=controller,
            ctrlopts=ctrlopts,
            backend=backend,
            cache=cache,
            dataarg=dataarg,
            dataopts=dataopts,
            metadir=metadir,
            accept_metadir=accept_metadir,
            modelsetup=modelsetup,
            modelopts=modelopts,
            updateinterval=updateinterval,
            loginterval=loginterval,
            visualize=visualize,
            strategy=strategy,
            strategyopts=strategyopts,
        )
        rc = RC_SUCCEEDED
    except:
        if rc != RC_SUCCEEDED:
            if logging.root.level < logging.INFO:
                log.exception("workflow failed")
            exc = click.exceptions.ClickException(
                click.style("Workflow failed: {}".format(sys.exc_info()[1]), fg="red")
            )
            exc.exit_code = rc
            raise exc


if __name__ == "__main__":
    main()



import copy
import hashlib
import json
import logging
import os
import uuid

import jq
import jsonpointer
import yaml

from yadageschemas.utils import WithJsonRefEncoder


class outputReference(object):
    def __init__(self, stepid, pointer):
        self.stepid = stepid
        self.pointer = pointer

    def __repr__(self):
        return "outputReference {}#{}".format(self.stepid, self.pointer.path)

    # (de-)serialization
    @classmethod
    def fromJSON(cls, data):
        return cls(data["stepid"], jsonpointer.JsonPointer(data["pointer_path"]))

    def json(self):
        return {"stepid": self.stepid, "pointer_path": self.pointer.path}


log = logging.getLogger(__name__)

DEFAULT_ID_METHOD = "uuid"


def json_hash(jsonable):
    """
    returns a content-based hash of a json-able structure (serialized with WithJsonRefEncoder)

    :param jsonable: a json-serializable object
    :return: the hash
    """
    the_hash = hashlib.sha1(
        json.dumps(jsonable, cls=WithJsonRefEncoder, sort_keys=True).encode("utf-8")
    ).hexdigest()
    return the_hash


def get_id_fromjson(jsonobject, method=DEFAULT_ID_METHOD):
    method = os.environ.get("YADAGE_ID_METHOD", method)
    if method == "uuid":
        return str(uuid.uuid4())
    elif method == "jsonhash":
        return json_hash(jsonobject)
    else:
        raise NotImplementedError("unkown id generation method {}".format(method))


def get_obj_id(obj_with_json_method, method=DEFAULT_ID_METHOD):
    return get_id_fromjson(obj_with_json_method.json(), method)


def process_refs(x, dag):
    if type(x) == dict:
        for k, v in x.items():
            x[k] = process_refs(v, dag)
        return x
    elif type(x) == list:
        for i, e in enumerate(x):
            x[i] = process_refs(e, dag)
        return x
    elif type(x) == outputReference:
        return x.pointer.resolve(dag.getNode(x.stepid).result)
    else:
        return x


def leaf_iterator_jsonlike(jsonlike, path=None):
    """'
    :param jsonlike: a jsonlike object (i.e. nested lists/arrays: leaf-types must not be JSONable)
    :return: iterator that yields (JSON path, leaf value) tuples
    """
    path = path or []
    if type(jsonlike) == list:
        for i, x in enumerate(jsonlike):
            thispath = copy.deepcopy(path)
            thispath.append(i)
            for leaf in leaf_iterator_jsonlike(x, path=thispath):
                yield leaf
    elif type(jsonlike) == dict:
        for k, v in jsonlike.items():
            thispath = copy.deepcopy(path)
            thispath.append(k)
            for leaf in leaf_iterator_jsonlike(v, path=thispath):
                yield leaf
    else:
        yield jsonpointer.JsonPointer.from_parts(path), jsonlike


def process_jsonlike(jsonlike, jq_obj_selector, callback):
    wflowrefs = [
        jsonpointer.JsonPointer.from_parts(x[1:])
        for x in jq.jq(
            "paths(if objects then {} else false end)".format(jq_obj_selector)
        ).transform({"value": jsonlike}, multiple_output=True)
    ]
    for wflowref in wflowrefs:
        value = callback(wflowref.resolve(jsonlike))
        if wflowref.path == "":
            return value
        else:
            wflowref.set(jsonlike, value)
    return jsonlike


def options_from_eqdelimstring(opts):
    options = {}
    for x in opts:
        key, value = x.split("=", 1)
        options[key] = yaml.safe_load(value)
    return options


def getinit_data(initfiles, parameters):
    """
    get initial data from both a list of files and a list of 'pname=pvalue'
    strings as they are passed in the command line <pvalue> is assumed to be a
    YAML parsable string.
    """

    initdata = {}
    for initfile in initfiles:
        log.debug("loading initialization data from file %s", initfile)
        initdata.update(**yaml.safe_load(open(initfile)))

    initdata.update(**options_from_eqdelimstring(parameters))
    return initdata


def setupbackend_fromstring(backend, backendopts=None):
    backendopts = backendopts or {}
    log.debug("setting up backend %s with opts %s", backend, backendopts)
    import yadage.backends.packtivitybackend as pb

    return pb.PacktivityBackend(
        packtivity_backendstring=backend, backendopts=backendopts
    )


def init_stage_spec(
    parameters, discover, used_inputs, name, nodename=None, relative=False
):
    return {
        "name": name,
        "dependencies": {"dependency_type": "jsonpath_ready", "expressions": []},
        "scheduler": {
            "scheduler_type": "init-stage",
            "parameters": parameters,
            "inputs": used_inputs,
            "nodename": nodename,
            "step": {
                "process": None,
                "environment": None,
                "publisher": {
                    "publisher_type": "fromparjq-pub",
                    "script": ".",
                    "tryExact": True,
                    "glob": discover,
                    "relative_paths": relative,
                },
            },
        },
    }


def stages_in_scope(workflow, scope):
    return jsonpointer.JsonPointer(scope).resolve(workflow.bookkeeping)["_meta"][
        "stages"
    ]


def rule_steps_indices(workflow):
    rule_to_steps_index = {}
    steps_to_rule_index = {}
    rule_to_subscopes_index = {}
    for rule in workflow.rules + workflow.applied_rules:
        path = "/".join([rule.offset, rule.rule.name])
        p = jsonpointer.JsonPointer(path)
        try:
            steps_of_rule = [
                x["_nodeid"] for x in p.resolve(workflow.stepsbystage) if "_nodeid" in x
            ]
        except jsonpointer.JsonPointerException:
            steps_of_rule = []

        try:
            a = p.resolve(workflow.stepsbystage)
            subscopes_of_rule = [
                # ['{}/{}'.format(x['_offset'],substage) for substage in x.keys() if not substage== '_offset' ]
                x["_offset"]
                for x in a
                if "_offset" in x
            ]
        except jsonpointer.JsonPointerException:
            subscopes_of_rule = []

        rule_to_steps_index[rule.identifier] = steps_of_rule
        rule_to_subscopes_index[rule.identifier] = subscopes_of_rule
        for step in steps_of_rule:
            steps_to_rule_index[step] = rule.identifier
    return rule_to_steps_index, steps_to_rule_index, rule_to_subscopes_index


def advance_coroutine(coroutine):
    try:
        return coroutine.next()
    except AttributeError:
        return coroutine.__next__()


def prepare_meta(metadir, accept=False):
    """
    prepare workflow meta-data directory

    :param metadir: the meta-data directory name
    :param accept: whether to accept an existing metadata directory
    """
    if os.path.exists(metadir):
        if not accept:
            raise RuntimeError(
                "yadage meta directory %s exists. Allow overwrite by using the command line option  --accept-metadir"
                % metadir
            )
    else:
        os.makedirs(metadir)


def pointerize(data, asref=False, stepid=None):
    def callback(p):
        return (
            outputReference(stepid, p)
            if asref
            else {"$wflowpointer": {"step": stepid, "result": p.path}}
            if stepid
            else p.path
        )

    return data.asrefs(callback=callback)



import json
import logging
import os
import time

import yadage.backends.federatedbackend as federatedbackend
from packtivity import datamodel as _datamodel

from packtivity.statecontexts import load_state
from yadage.utils import json_hash

from ..backends import CachedProxy
from .trivialbackend import TrivialBackend, TrivialProxy

log = logging.getLogger(__name__)

from ..handlers.utils import handler_decorator

cachehandlers, cachestrategy = handler_decorator()


@cachestrategy("checksums")
def checksum_cache(config):
    configfile = config
    return ChecksumCache(configfile)


def setupcache_fromstring(configstring):
    """
    generate cache from a string configuration (such as those passed from CLIs).
    valid strings are:

    - checksums:<path to cache file (JSON)>

    :param configstring: the configuration string
    :return: a cache object
    """
    configparts = configstring.split(":", 1)
    strategy, specificconf = configparts
    try:
        return cachehandlers[strategy](specificconf)
    except KeyError:
        raise RuntimeError("unknown caching config")


class CachedBackend(federatedbackend.FederatedBackend):
    """
    A caching backend that takes an existing backend and a cache configuration.
    If at submission time a cache for a given task is valid, the result will be
    returned directly via a trivial proxy. If not submission proceeds as normal
    """

    def __init__(self, backend, cache):
        super(CachedBackend, self).__init__(
            {"cache": TrivialBackend(), "primary": backend}
        )
        self.cache = cache
        self.primary_enabled = True

    def ready(self, proxy):
        isready = super(CachedBackend, self).ready(proxy)
        # if this a genuinely new result, cache it under the task cache id
        if isready and type(proxy) is not TrivialProxy:
            result = super(CachedBackend, self).result(proxy)
            status = super(CachedBackend, self).successful(proxy)
            if not self.cache.cacheexists(proxy.cacheid):
                self.cache.cacheresult(proxy.cacheid, status, result)
        return isready

    def expected_result(self, task):
        return self.backends["packtivity"].prepublish(task)

    def routedsubmit(self, task):
        cached = self.cache.cacheddata(task)
        if cached:
            log.info("use cached result for task: %s", task.metadata["name"])
            return TrivialProxy(
                cached["status"], cached["result"].json(), task.state.datamodel
            )
        else:
            if not self.primary_enabled:
                raise RuntimeError(
                    "cache failed but refusing to submit to primary since it was explicitly disabled"
                )
            log.info("do proper submit for task: %s", task.metadata["name"])
            # create id for this task using the cache builder, with which
            # we will store the result with once it's ready
            cacheid = self.cache.cacheid(task)

            task.state.reset()
            # maybe we erroneously are re-submitting this see
            # https://github.com/diana-hep/yadage/issues/45 TODO!
            primaryproxy = self.backends["primary"].submit(
                task.spec, task.parameters, task.state, task.metadata
            )
            cachedproxy = CachedProxy(primaryproxy, cacheid)
            return cachedproxy

    def routeproxy(self, proxy):
        if type(proxy) == TrivialProxy:
            return "cache", proxy
        else:
            return "primary", proxy.proxy


class CacheBuilder(object):
    """
    Cache base class. Implementations need to provide cachevalid and
    generate_validation_data methods
    """

    def __init__(self, cachefile):
        self.datamodel = _datamodel
        self.cachefile = cachefile
        if not os.path.exists(self.cachefile):
            log.info("fresh cache as file at %s does not exist yet.", self.cachefile)
            self.cache = {}
        else:
            log.info("reading cache from %s", self.cachefile)
            self.cache = json.load(open(self.cachefile))

    def todisk(self):
        log.debug("writing cache to %s", self.cachefile)
        json.dump(self.cache, open(self.cachefile, "w"), indent=4, sort_keys=True)

    def remove(self, cacheid):
        self.cache.pop(cacheid)

    def cacheresult(self, cacheid, status, result):
        """
        saves a result and its status under a unique identifier and includes
        validation data to verify the cache validity at a later point in time
        """
        validation_data = self.generate_validation_data(cacheid)
        log.debug("caching result for cacheid: %s", cacheid)
        log.debug(
            "caching result for process: %s",
            self.cache[cacheid]["task"]["spec"]["process"],
        )
        log.debug("caching with validation data %s", validation_data)
        self.cache[cacheid]["result"] = {
            "status": "SUCCESS" if status else "FAILED",
            "result": result.json(),
            "cachingtime": time.time(),
            "validation_data": validation_data,
        }
        self.todisk()

    def cachedresult(self, cacheid, state, silent=True):
        """
        returns the cached result. when silent = True the mthod exits gracefully
        and returns None
        """
        if silent:
            if not self.cacheexists(cacheid):
                return None
        return {
            "result": self.datamodel.create(
                self.cache[cacheid]["result"]["result"], state.datamodel
            ),
            "status": self.cache[cacheid]["result"]["status"],
        }

    def cacheid(self, task):
        t = task.json()
        tocache = [t["state"], t["parameters"], t["spec"]]
        hash = json_hash(tocache)
        log.debug("%s -> %s", json.dumps(tocache), hash)
        return hash

    def cacheexists(self, cacheid):
        return cacheid in self.cache and "result" in self.cache[cacheid]

    def cacheddata(self, task, remove_invalid=True):
        """
        returns results from a valid cache entry if it exists, else None
        if remove_invalid = True, also removes invalid cache entries from cache
        """
        cacheid = self.cacheid(task)
        # register this task with the cacheid if we don't know about it yet
        log.debug("checking cache for task %s %s", task.metadata["name"], cacheid)
        if cacheid not in self.cache:
            self.cache[cacheid] = {"task": task.json()}
        if not self.cacheexists(cacheid):
            log.info(
                "cache non-existent for task %s (%s)", task.metadata["name"], cacheid
            )
            return None
        if not self.cachevalid(cacheid):
            self.remove(cacheid)
            self.cache[cacheid] = {"task": task.json()}
            return None
        # return a cached result if we have one if not, return None
        result = self.cachedresult(cacheid, task.state, silent=True)
        log.debug("returning cached result %s", result)
        return result

    def cachevalid(self, cacheid):
        raise NotImplementedError

    def generate_validation_data(self, cacheid):
        raise NotImplementedError


class ChecksumCache(CacheBuilder):
    """
    This is a cache that based on state hashes
    """

    def __init__(self, cachefile):
        super(ChecksumCache, self).__init__(cachefile)

    def remove(self, cacheid):
        """
        removes entry from the cache and reset its state

        :param cachid: the cache entry identifier
        """
        task = self.cache[cacheid]["task"]
        log.debug("removing cache entry %s and resetting state", cacheid)
        load_state(task["state"]).reset()
        super(ChecksumCache, self).remove(cacheid)

    def generate_validation_data(self, cacheid):
        """
        calculate validation data (checksums in SHA1 hash form)

        :param cachid: the cache entry identifier
        """
        validation_data = {
            "state_hash": load_state(self.cache[cacheid]["task"]["state"]).state_hash()
        }

        log.debug("validation data is %s", validation_data)
        return validation_data

    def cachevalid(self, cacheid):
        """
        validate cache entry

        :param cachid: the cache entry identifier
        """
        # task = self.cache[cacheid]['task']

        stored_validation_data = self.cache[cacheid]["result"]["validation_data"]
        validation_data_now = self.generate_validation_data(cacheid)

        if not stored_validation_data == validation_data_now:
            log.info(
                "cache invalid stored data:\n%s\ncurrent data:\n%s",
                stored_validation_data,
                validation_data_now,
            )
            return False
        return True



import logging
import shlex
import subprocess

import jsonpointer
import pydotplus

log = logging.getLogger(__name__)


def fillscope(cluster, workflow, nodes_to_connect, scope="", subcluster=True):
    # scopecluster = stagecluster = pydotplus.graphviz.Cluster(graph_name =
    # '_'.join(stagescopeprts),

    if subcluster:
        scopecluster = pydotplus.graphviz.Cluster(
            graph_name=scope.replace("/", ""),
            label="".join(["[{}]".format(p) for p in scope.split("/")[1:]]),
            style="solid",
            color="blue",
        )
        cluster.add_subgraph(scopecluster)
    else:
        scopecluster = cluster
    scopeptr = jsonpointer.JsonPointer(scope)
    scoped = scopeptr.resolve(workflow.stepsbystage)
    for stage, elements in scoped.items():
        stagescopeprts = scopeptr.parts + [stage]

        if subcluster:
            stagecluster = pydotplus.graphviz.Cluster(
                graph_name="_".join(stagescopeprts),
                label=stage,
                labeljust="l",
                color="grey",
                style="dashed",
            )
            scopecluster.add_subgraph(stagecluster)
        else:
            stagecluster = scopecluster

        for i, element in enumerate(elements):
            if "_nodeid" in element:
                nodes_to_connect.append(element["_nodeid"])
                element = element["_nodeid"]
                targetcl = stagecluster if stage != "init" else scopecluster
                shape = "diamond" if stage in ["init", "output"] else "box"
                label = "" if stage in ["init", "output"] else "{}[{}]".format(stage, i)
                additional = (
                    {"fixedsize": True, "height": 0.2, "width": 0.2}
                    if stage in ["init", "output"]
                    else {}
                )
                nodeobj = workflow.dag.getNode(element)
                targetcl.add_node(
                    pydotplus.graphviz.Node(
                        element,
                        label=label,
                        color=("blue" if nodeobj.ready() else "grey"),
                        shape=shape,
                        **additional
                    )
                )
                result = nodeobj.result if nodeobj.has_result() else None
                if result:
                    add_result(targetcl, element, result, ready=nodeobj.ready())
            elif type(element) == dict:
                # recurse...
                fillscope(
                    stagecluster,
                    workflow,
                    nodes_to_connect,
                    jsonpointer.JsonPointer.from_parts(
                        scopeptr.parts + [stage, i]
                    ).path,
                    subcluster=subcluster,
                )


def path_to_id(stepid, path):
    return "{}_{}".format(stepid, path.replace("/", "_"))


def add_result(graph, parent, data, ready):
    leafpointers = [p for p, v in data.leafs()]

    for leaf in leafpointers:
        leafid = path_to_id(parent, leaf.path)
        # value = leaf.resolve(jsondata)
        source = "".join(["[{}]".format(p) for p in leaf.parts])
        label = "{}".format(source)
        graph.add_node(
            pydotplus.graphviz.Node(
                leafid, label=label, color=("red" if ready else "grey")
            )
        )
        graph.add_edge(pydotplus.graphviz.Edge(parent, leafid))


def attach_to_results(provgraph, workflow, node):
    nodeobj = workflow.dag.getNode(node)
    for dep in nodeobj.task.inputs:
        if workflow.dag.getNode(dep.stepid).has_result():
            resultid = path_to_id(dep.stepid, dep.pointer.path)
            provgraph.add_edge(pydotplus.graphviz.Edge(resultid, node))


def connect_nodes(provgraph, workflow, nodes):
    for node in nodes:
        attach_to_results(provgraph, workflow, node)


def provdotgraph(workflow, subcluster=True, scope=""):
    provgraph = pydotplus.graphviz.Graph()
    nodes_to_connect = []
    fillscope(provgraph, workflow, nodes_to_connect, scope=scope, subcluster=subcluster)
    connect_nodes(provgraph, workflow, nodes_to_connect)
    return provgraph


def write_prov_graph(workdir, workflow, scope="", vizformat="pdf"):
    provgraph = provdotgraph(workflow, subcluster=True, scope=scope)

    with open("{}/yadage_workflow_instance.dot".format(workdir), "w") as dotfile:
        dotfile.write(provgraph.to_string())

    with open(
        "{}/yadage_workflow_instance.{}".format(workdir, vizformat), "w"
    ) as outfile:
        subprocess.call(
            shlex.split(
                "dot -T{} {}/yadage_workflow_instance.dot".format(vizformat, workdir)
            ),
            stdout=outfile,
        )



import logging

import networkx as nx

import yadage.utils as utils
import adage.nodestate

log = logging.getLogger(__name__)


def reset_node_state(node):
    node.submit_time = None
    node.ready_by_time = None
    node.resultproxy = None
    node.backend = None
    node.expected_result = None
    node._result = None
    node._state = adage.nodestate.DEFINED


def reset_step(workflow, step):
    s = workflow.dag.getNode(step)
    log.debug(
        "resetting DAG node %s/%s:%s",
        s.task.metadata["wflow_offset"],
        s.task.metadata["wflow_stage"],
        s.task.metadata["wflow_stage_node_idx"],
    )
    reset_node_state(s)
    try:
        s.task.state.reset()
    except AttributeError:
        pass


def remove_rule(workflow, ruleid):
    assert ruleid in [x.identifier for x in workflow.rules]
    r = workflow.view().getRule(identifier=ruleid)
    log.debug(">>> removing rule %s/%s", r.offset, r.rule.name)
    workflow.rules.remove(r)
    view = workflow.view(r.offset)
    stagemeta = view.bookkeeper["_meta"]["stages"]
    view.bookkeeper["_meta"]["stages"] = [x for x in stagemeta if not x == ruleid]


def remove_rules(workflow, ruleids):
    for r in ruleids:
        remove_rule(workflow, r)


def undo_rule(workflow, ruleid):
    r2s, s2r, r2subscopes = utils.rule_steps_indices(workflow)

    if not ruleid in [r.identifier for r in workflow.applied_rules]:
        ruleobj = workflow.view().getRule(identifier=ruleid)
        log.debug(
            "rule %s/%s not in list of applied rules. possibly already undone during recursion.",
            ruleobj.offset,
            ruleobj.rule.name,
        )
        return

    downstream_nodes_rules = list(
        set([s2r[s] for s in collective_downstream(workflow, r2s[ruleid])])
    )

    r = workflow.view().getRule(identifier=ruleid)
    log.debug(
        "undoing %s/%s. downstream_nodes_rules: %s",
        r.offset,
        r.rule.name,
        len(downstream_nodes_rules),
    )

    assert ruleid not in downstream_nodes_rules
    # TODO: not an issue now, but if stages schedule linked nodes we might
    # we might see downstream steps that are part of the same rule

    if not downstream_nodes_rules:
        idx = workflow.applied_rules.index(r)
        stepids = r2s[ruleid]
        steps = [workflow.dag.getNode(nid) for nid in stepids]

        # reset states of the step
        reset_steps(workflow, stepids)
        #
        # delete the steps
        for n in steps:
            log.debug(
                "remove DAG node %s/%s:%s",
                n.task.metadata["wflow_offset"],
                n.task.metadata["wflow_stage"],
                n.task.metadata["wflow_stage_node_idx"],
            )
            workflow.dag.removeNode(n)

        for subscope in r2subscopes[ruleid]:
            subrules = utils.stages_in_scope(workflow, subscope)
            log.debug("remove subscope %s with %s subrules", subscope, len(subrules))
            for subruleidx, subrule in enumerate(subrules):
                subruleobj = workflow.view().getRule(identifier=subrule)
                log.debug(
                    "undo sub DAG rule %s/%s", subruleobj.offset, subruleobj.rule.name
                )
                undo_rule(workflow, subrule)

            log.debug("removing %s subrules", len(subrules))
            for subruleidx, subrule in enumerate(subrules):
                subruleobj = workflow.view().getRule(identifier=subrule)
                log.debug(
                    "remove sub DAG rule %s/%s", subruleobj.offset, subruleobj.rule.name
                )
                remove_rule(workflow, subrule)
        # remove rule from applied list
        r = workflow.applied_rules.pop(idx)

        # reset all index data for stage
        workflow.view(r.offset).steps.pop(r.rule.name)

        # remove steps and stage from bookkeeping
        for s in stepids:
            workflow.view(r.offset).bookkeeper["_meta"]["steps"].remove(s)

        workflow.view(r.offset).bookkeeper["_meta"]["stages"].remove(ruleid)

        # re-append the rule
        log.debug("re-appened {}/{}".format(r.offset, r.rule.name))

        # workflow.steps[stage][-1]['_offset'] = offset
        newid = workflow.view(r.offset).addRule(r.rule, identifier=r.identifier)
        assert newid == r.identifier

        log.debug("undo any rules that would not be applicable now")
        for rule in workflow.applied_rules:
            if not rule.applicable(workflow):
                ruleobj = workflow.view().getRule(identifier=rule.identifier)
                log.debug(
                    "rule would not be appilcable in current state so undo >> %s/%s",
                    ruleobj.offset,
                    ruleobj.rule.name,
                )
                undo_rule(workflow, rule.identifier)

        return

    log.debug(
        "to undo a rule, we need to undo rules responsible for any of the downstream nodes of the rules of this stage"
    )
    while True:
        if not downstream_nodes_rules:
            break
        for r in downstream_nodes_rules:
            log.debug("undoing a downstream rule")
            undo_rule(workflow, r)
            log.debug("undone a downstream rule")
        r2s, s2r, _ = utils.rule_steps_indices(workflow)
        downstream_nodes_rules = list(
            set([s2r[s] for s in collective_downstream(workflow, r2s[ruleid])])
        )
        log.debug(
            "re-asses if there are still any downstream rules: {}".format(
                len(downstream_nodes_rules)
            )
        )
    log.debug("undo rule (now without any downstream rules)")
    undo_rule(workflow, ruleid)


def undo_rules(workflow, ruleids):
    for r in ruleids:
        undo_rule(workflow, r)


def reset_steps(workflow, steps):
    for s in steps:
        reset_step(workflow, s)


def collective_downstream(workflow, steps):
    downstream = set()
    for step in steps:
        for x in nx.descendants(workflow.dag, step):
            downstream.add(x)
    return list(downstream)



import os

# import datetime
# import time

import jsonpointer

import adage.node
import adage.serialize

from packtivity import datamodel

from packtivity.backendutils import load_proxy
from .tasks import outputReference, packtivity_task


class YadageNode(adage.node.Node):
    """
    Node object for yadage that extends the default with
    the ability to have prepublished results
    """

    def __init__(self, name, task, identifier=None, result=None):
        super(YadageNode, self).__init__(name, task, identifier, result=result)
        self.expected_result = None

    def __repr__(self):
        # lifetime = datetime.timedelta(seconds = (time.time() - self.define_time))
        # runtime = None
        if self.state != adage.nodestate.DEFINED:
            pass
            # referencetime = time.time() if not self.ready() else self.ready_by_time
            # runtime = datetime.timedelta(seconds = (referencetime - self.submit_time))
        return "<{}/{}:{}|{}|{}>".format(
            self.task.metadata.get("wflow_offset", "-"),
            self.task.metadata.get("wflow_stage", "-"),
            self.task.metadata.get("wflow_stage_node_idx", "-"),
            str(self.state).lower(),
            "known" if self.has_result() else "unknown",
        )

    def has_result(self):
        return (self.expected_result is not None) or self.successful()

    @property
    def result(self):
        if (
            self.expected_result is not None
            and "YADAGE_IGNORE_PREPUBLISHING" not in os.environ
        ):
            if self.ready() and self.successful():
                sanity = (
                    super(YadageNode, self).result.json() == self.expected_result.json()
                )
                if not sanity:
                    raise RuntimeError(
                        "prepublished and actual result differ:\n result:\n{}\n prepub:{}".format(
                            super(YadageNode, self).result.json(),
                            self.expected_result.json(),
                        )
                    )
            return self.expected_result
        return super(YadageNode, self).result

    def readfromresult(self, pointerpath, trackinputs=None, failsilently=False):
        if not self.has_result():
            if failsilently:
                return None
            raise RuntimeError("attempt")
        pointer = jsonpointer.JsonPointer(pointerpath)
        if trackinputs is not None:
            trackinputs.append(outputReference(self.identifier, pointer))
        v = self.result.resolve_ref(pointer)
        return v

    def json(self):
        json_or_nil = lambda x: None if x is None else x.json()
        d = adage.serialize.node_to_json(self, json_or_nil, json_or_nil)
        d["result"] = json_or_nil(self.result if self.has_result() else None)
        return d

    @classmethod
    def fromJSON(cls, data, deserialization_opts=None):
        if data["task"]["type"] == "packtivity_task":
            task = packtivity_task.fromJSON(data["task"], deserialization_opts)
            result = (
                datamodel.create(data["result"], getattr(task.state, "datamodel", {}))
                if data["result"]
                else None
            )
            instance = cls(data["name"], task, data["id"], result)

            adage.serialize.set_generic_data(instance, data)
            instance.resultproxy = (
                load_proxy(
                    data["proxy"], deserialization_opts, best_effort_backend=False
                )
                if data["proxy"]
                else None
            )
            return instance
        else:
            raise RuntimeError("unknown task type", data["task"]["type"])



import logging

from .handlers.predicate_handlers import handlers as pred_handlers
from .utils import get_id_fromjson
from .state_providers import load_provider

from packtivity import datamodel as _datamodel

log = logging.getLogger(__name__)


class OffsetStage(object):
    """
    A wrapper object around a scoped rule, so that it can be applied from
    a global p.o.v., i.e. as adage expects its rules.
    """

    def __init__(self, rule, offset=None, identifier=None):
        """
        initializes a scoped rule. scope is defined by a JSONPointer, e.g.
        one[0]two.three

        """

        self.rule = rule
        self.offset = offset
        self.identifier = identifier or get_id_fromjson(
            {"rule": rule.json(), "offset": offset}
        )

    def __repr__(self):
        return "<OffsetStage {}/{} >".format(self.offset, self.rule.name)

    def applicable(self, adageobj):
        """
        determin whether the rule is applicable. Evaluated within the offset.
        :param adageobj: the workflow object
        """
        from .wflowview import WorkflowView  # importing here to avoid circdep

        x = self.rule.applicable(WorkflowView(adageobj, self.offset))
        return x

    def apply(self, adageobj):
        """
        applies a rule within the scope set by offset

        :param adageobj: the workflow object
        """
        from .wflowview import WorkflowView  # importing here to avoid circdep

        self.rule.apply(WorkflowView(adageobj, self.offset))

    # (de-)serialization
    @classmethod
    def fromJSON(cls, data, deserialization_opts=None):
        if data["rule"]["type"] == "JsonStage":
            rule = JsonStage.fromJSON(data["rule"], deserialization_opts)
        else:
            RuntimeError("unknown stage type %s", data["rule"]["type"])
        return cls(rule=rule, identifier=data["id"], offset=data["offset"])

    def json(self):
        return {
            "type": "offset",
            "id": self.identifier,
            "offset": self.offset,
            "rule": self.rule.json(),
        }


class ViewStageBase(object):
    """
    base class for workflow stages operating on workflow views, that may
    expose only a partial slice of an overall workflow.
    The class also provides common methods to ease modifying the workflow.

    Implementations are required to provide
    * a ready() method to implement the predicate method
    * a schedule() method that is called upon apply
    """

    def __init__(self, name, state_provider):
        self.name = name
        self.state_provider = state_provider
        self.datamodel = _datamodel
        self.view = None

    def schedule(self):
        raise NotImplementedError()

    def ready(self):
        raise NotImplementedError()

    def applicable(self, flowview):
        self.view = flowview
        return self.ready()

    def apply(self, flowview):
        self.view = flowview
        self.schedule()

    def addStep(self, step):
        dependencies = [self.view.dag.getNode(k.stepid) for k in step.inputs]
        return self.view.addStep(step, stage=self.name, depends_on=dependencies)

    def addWorkflow(self, rules, isolate=True):
        self.view.addWorkflow(rules, stage=self.name if isolate else None)

    def json(self):
        return {
            "name": self.name,
            "state_provider": self.state_provider.json()
            if self.state_provider
            else None,
        }


class JsonStage(ViewStageBase):
    """
    A stage that is defined via the JSON scheduler schemas
    """

    def __init__(self, json, state_provider):
        self.stagespec = json["scheduler"]
        self.depspec = json["dependencies"]
        super(JsonStage, self).__init__(json["name"], state_provider)

    def __repr__(self):
        return "<JsonStage: {}>".format(self.name)

    def ready(self):
        if not self.depspec:
            return True
        predicate = pred_handlers[self.depspec["dependency_type"]]
        return predicate(self, self.depspec, self.stagespec)

    def schedule(self):
        # imported here to avoid circular dependency
        log.debug("scheduling")
        from .handlers.scheduler_handlers import handlers as sched_handlers

        scheduler = sched_handlers[self.stagespec["scheduler_type"]]
        scheduler(self, self.stagespec)
        log.debug("scheduled")

    # (de-)serialization
    @classmethod
    def fromJSON(cls, data, deserialization_opts=None):
        return cls(
            json={
                "scheduler": data["scheduler"],
                "name": data["name"],
                "dependencies": data["dependencies"],
            },
            state_provider=load_provider(data["state_provider"], deserialization_opts),
        )

    def json(self):
        data = super(JsonStage, self).json()
        data.update(
            type="JsonStage", scheduler=self.stagespec, dependencies=self.depspec
        )
        return data



import yadageschemas


def workflow(
    source,
    toplevel,
    schema_name="yadage/workflow-schema",
    schemadir=None,
    validate=True,
):
    """
    load (and validate) workflow from source
    param source: URI fragment to the source
    param toplevel: base URI to resolve references from
    schemadir
    """
    dialect, spec = (
        ("raw_with_defaults", source) if ":" not in source else source.split(":", 1)
    )

    specopts = {
        "toplevel": toplevel,
        "schema_name": schema_name,
        "schemadir": schemadir,
        "load_as_ref": False,
    }
    validopts = {"schema_name": schema_name, "schemadir": schemadir}

    data = yadageschemas.load(
        spec, specopts, validate=True, validopts=validopts, dialect=dialect
    )
    return data


def validate(data, schema_name="yadage/workflow-schema", schemadir=None):
    """validate workflow data"""
    validopts = {"schema_name": schema_name, "schemadir": schemadir}
    yadageschemas.validate_spec(data, validopts)



class FederatedBackend(object):
    """
    A meta backend that routes tasks to several internal backends
    """

    def __init__(self, backends):
        """takes a dictionary of backendname->backendobject and a router"""
        self.backends = backends

    def routedsubmit(self, task):
        raise NotImplementedError("needs implementation")

    def routedbatchsubmit(self, task):
        raise NotImplementedError("needs implementation")

    def routeproxy(self, proxy):
        raise NotImplementedError("needs implementation")

    def batch_submit(self, tasks):
        return self.routedbatchsubmit(tasks)

    def submit(self, task):
        return self.routedsubmit(task)

    def result(self, proxy):
        b, p = self.routeproxy(proxy)
        return self.backends[b].result(p)

    def expected_result(self, proxy):
        b, p = self.routeproxy(proxy)
        return self.backends[b].expected_result(p)

    def ready(self, proxy):
        b, p = self.routeproxy(proxy)
        return self.backends[b].ready(p)

    def successful(self, proxy):
        b, p = self.routeproxy(proxy)
        return self.backends[b].successful(p)

    def fail_info(self, proxy):
        b, p = self.routeproxy(proxy)
        return self.backends[b].fail_info(p)



#!/usr/bin/env python
import functools
import logging
import os

import click
import yaml

import yadage.manualutils as manualutils
import yadage.utils as utils
import yadage.workflow_loader as workflow_loader

from .steering_api import execute_steering
from .steering_object import YadageSteering
from .logutils import LOGFORMAT

log = logging.getLogger(__name__)


def connection_options(func):
    @click.option(
        "-m",
        "--metadir",
        default="yadagemeta",
        help="directory to store workflow metadata",
    )
    @click.option("--accept-metadir/--no-accept-metadir", default=True)
    @click.option("-r", "--controller", default="frommodel")
    @click.option(
        "-o",
        "--ctrlopt",
        multiple=True,
        default=None,
        help="options for the workflow controller",
    )
    @click.option("-s", "--modelsetup", default="filebacked")
    @click.option(
        "-l",
        "--modelopt",
        multiple=True,
        default=None,
        help="options for the workflow state models",
    )
    @click.option("-b", "--backend", default="foregroundasync")
    @click.option("--local/--remote", default=True)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def common_options(func):
    @click.option("-v", "--verbosity", default="INFO")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.group()
def mancli():
    pass


@mancli.command()
@click.option("-s", "--modelsetup", default="filebacked:yadagemeta/yadage_state.json")
@click.option("-t", "--toplevel", default=os.getcwd())
@click.option(
    "-d",
    "--dataopt",
    multiple=True,
    default=None,
    help="options for the workflow data state",
)
@click.option(
    "--metadir", default="yadagemeta", help="directory to store workflow metadata"
)
@click.option("--parameter", "-p", multiple=True)
@click.argument("dataarg", default="workdir")
@click.argument("workflow", default="workflow.yml")
@click.argument("initfiles", nargs=-1)
def init(
    dataarg, workflow, initfiles, modelsetup, dataopt, metadir, toplevel, parameter
):
    if os.path.exists("input.yml") and not initfiles:
        initfiles = ("input.yml",)
    initdata = utils.getinit_data(initfiles, parameter)
    dataopts = utils.options_from_eqdelimstring(dataopt)

    ys = YadageSteering.create(
        dataarg=dataarg,
        dataopts=dataopts,
        workflow=workflow,
        toplevel=toplevel,
        initdata=initdata,
        metadir=metadir,
        modelsetup=modelsetup,
    )
    assert ys


def click_print_applicable_stages(controller):
    click.secho("Applicable Stages: ", fg="blue")
    for x in controller.adageobj.rules:
        if x.identifier in controller.applicable_rules():
            click.secho("{}/{}".format(x.offset, x.rule.name))


def click_print_submittable_nodes(controller):
    click.secho("Submittable Nodes: ", fg="blue")
    _, s2r, _ = utils.rule_steps_indices(controller.adageobj)
    for x in controller.adageobj.dag.nodes():
        node = controller.adageobj.dag.getNode(x)
        rule = controller.adageobj.view().getRule(identifier=s2r[node.identifier])
        if node.identifier in controller.submittable_nodes():
            click.secho(
                "node: {} ({}) part of stage {}".format(
                    node.name, node.identifier, "/".join([rule.offset, rule.rule.name])
                )
            )


def click_print_rule(rule, step_index, subrule_index, step_status):
    steps_of_rule = step_index[rule.identifier]
    substages_of_rule = subrule_index[rule.identifier]
    rule_stats = {}
    for x in steps_of_rule:
        rule_stats.setdefault(step_status[x], 0)
        rule_stats[step_status[x]] += 1
    click.secho(
        "{}/{} [steps: ({})] [subflows: {}]".format(
            rule.offset,
            rule.rule.name,
            "/".join("{}: {}".format(k, v) for k, v in rule_stats.items()),
            len(substages_of_rule),
        )
    )


def click_print_applied_stages(controller):
    click.secho("Applied Stages: ", fg="blue")
    r2s, s2r, r2sub = utils.rule_steps_indices(controller.adageobj)
    step_status = {s: str(controller.adageobj.dag.getNode(s).state) for s in s2r.keys()}
    # print(step_status)
    for x in sorted(
        controller.adageobj.applied_rules,
        key=lambda r: "{}/{}".format(r.offset, r.rule.name),
    ):
        click_print_rule(x, r2s, r2sub, step_status)


@mancli.command()
@click.argument("name", default=None, nargs=-1)
@connection_options
@common_options
@click.option("--submit/--no-submit")
def apply_stage(
    name,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
    submit,
):

    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    if not name:
        click_print_applicable_stages(controller)
        return

    for n in name:
        offset, scopedname = n.rsplit("/", 1)
        rule = controller.adageobj.view(offset).getRule(scopedname)
        if not rule:
            click.secho(
                "No such stage {}, pick one of the applicable below:".format(n),
                fg="red",
            )
            click_print_applicable_stages(controller)
            return

        if rule.identifier in [r.identifier for r in controller.adageobj.applied_rules]:
            click.secho("Stage {} was already applied.".format(n), fg="yellow")
            continue

        if rule.identifier not in controller.applicable_rules():
            click.secho("Rule {} not yet applicable".format(n), fg="red")
            continue

        controller.apply_rules([rule.identifier])

        if submit:
            _, s2r, _ = utils.rule_steps_indices(controller.adageobj)
            nodes_to_submit = [
                x for x in controller.submittable_nodes() if s2r[x] == rule.identifier
            ]
            controller.submit_nodes(nodes_to_submit)

        click.secho("Stage {} applied".format(n), fg="green")


@mancli.command()
@click.option("-n", "--nodeid", default=None)
@click.option("-a", "--allof", default=None)
@click.option("-o", "--offset", default="")
@connection_options
@common_options
def submit(
    nodeid,
    allof,
    offset,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    if not (allof or nodeid):
        click_print_submittable_nodes(controller)
        return

    if nodeid:
        nodes_to_submit = [nodeid] if nodeid in controller.submittable_nodes() else []
    if allof:
        offset, name = allof.rsplit("/", 1)
        rule = controller.adageobj.view().getRule(name=name, offset=offset)
        if not rule:
            click.secho("stage not found!", fg="red")
            return

        all_submittable = controller.submittable_nodes()
        _, s2r, _ = utils.rule_steps_indices(controller.adageobj)
        nodes_to_submit = [x for x in all_submittable if s2r[x] == rule.identifier]

    if not nodes_to_submit:
        click.secho("No nodes to submit (perhaps already submitted?)", fg="yellow")
        return

    controller.submit_nodes(nodes_to_submit)
    click.secho("submitted: {}".format(nodes_to_submit), fg="green")


@mancli.command()
@connection_options
@common_options
def shell(
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    click.secho("launching yadage shell", fg="green")
    assert ys
    import IPython

    IPython.embed()


@mancli.command()
@connection_options
@common_options
@click.option("--show-processed/--hide-processed", default=False)
def show(
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
    show_processed,
):

    ctrlarg = controller  # for later
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    click.secho(
        """\
Workflow:
---------
state source: {statesource}
successful: {successful}
finished: {finished}
valid: {valid}
# of applicable rules: {applicable}
# of submittable nodes: {submittable}
""".format(
            statesource=modelsetup if local else "remote via {}".format(ctrlarg),
            successful=click.style(
                str(controller.successful()),
                fg="green" if controller.successful() else "red",
            ),
            finished=click.style(
                str(controller.finished()),
                fg="green" if controller.finished() else "yellow",
            ),
            valid=click.style(
                str(controller.validate()),
                fg="green" if controller.validate() else "red",
            ),
            applicable=len(controller.applicable_rules()),
            submittable=len(controller.submittable_nodes()),
        )
    )
    click_print_applicable_stages(controller)
    click_print_submittable_nodes(controller)

    if show_processed:
        click_print_applied_stages(controller)


@mancli.command()
@click.argument("name")
@connection_options
@common_options
def preview(
    name,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):

    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    preview = manualutils.preview_rule(controller.adageobj, name)
    if not preview:
        click.secho("cannot preview {}".format(name))
        raise click.Abort()

    new_rules, new_nodes = preview
    click.secho(
        "Preview of Stage: # new rules: {} # new nodes {}".format(
            len(new_rules), len(new_nodes)
        )
    )
    for n in new_nodes:
        click.secho(
            '-> new node "{}" with {} upstream dependencies'.format(
                n["name"], len(n["parents"])
            )
        )


def handle_common_options(verbosity):
    logging.basicConfig(level=getattr(logging, verbosity), format=LOGFORMAT)


def handle_connection_options(
    metadir, accept_metadir, controller, ctrlopt, modelsetup, modelopt, backend, local
):
    ctrlopts = utils.options_from_eqdelimstring(ctrlopt)
    modelopts = utils.options_from_eqdelimstring(modelopt)

    if modelsetup == "filebacked":
        modelsetup = "filebacked:{}/yadage_state.json".format(metadir)

    ys = manualutils.connect(
        metadir=metadir,
        accept_metadir=accept_metadir,
        ctrlstring=controller,
        ctrlopts=ctrlopts,
        modelsetup=modelsetup,
        modelopts=modelopts,
        backendstring=backend if local else None,
    )
    ys.controller.sync_backend()
    return ys


@mancli.command()
@click.option(
    "-n",
    "--nsteps",
    default=-1,
    help="number of steps to process. use -1 to for no limit (will run workflow to completion)",
)
@click.option("--track/--no-track", default=False)
@click.option("-u", "--update-interval", default=1)
@click.option("-g", "--strategy", help="set execution stragegy")
@click.option(
    "-y", "--strategyopt", help="strategy option", multiple=True, default=None
)
@connection_options
@common_options
def step(
    track,
    strategy,
    nsteps,
    update_interval,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    strategyopt,
    verbosity,
):

    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )

    strategyopts = utils.options_from_eqdelimstring(strategyopt)
    execute_steering(
        ys,
        updateinterval=update_interval,
        default_trackers=track,
        strategy=strategy,
        strategyopts=strategyopts,
    )


@mancli.command()
@click.argument("dataarg")
@click.argument("workflow")
@click.option(
    "-d", "--dataopts", multiple=True, default=None, help="ctrl opts for state provider"
)
@click.option("-f", "--offset", default="")
@click.option("-g", "--groupname", default=None)
@click.option("-t", "--toplevel", default=os.curdir)
@click.option("--parameter", "-p", multiple=True)
@connection_options
@common_options
def add(
    offset,
    groupname,
    dataarg,
    dataopts,
    workflow,
    toplevel,
    parameter,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    parameter = utils.options_from_eqdelimstring(parameter)
    dataopts = utils.options_from_eqdelimstring(dataopts)
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    stages = []
    if parameter:
        inspec = utils.init_stage_spec(parameter, False, [], "init")
        stages.append(inspec)

    stages = (
        stages
        + workflow_loader.workflow(workflow, toplevel=toplevel, validate=True)["stages"]
    )
    controller.add_rules(
        stages, dataarg, offset=offset, groupname=groupname, dataopts=dataopts
    )


@mancli.command()
@click.option("-f", "--fileformat", default="pdf")
@click.option("-w", "--workdir", default=os.curdir)
@connection_options
@common_options
def visualize(
    workdir,
    fileformat,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    from .visualize import write_prov_graph

    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller
    if controller.backend:
        controller.sync_backend()

    write_prov_graph(workdir, controller.adageobj, vizformat=fileformat)


@mancli.command()
@click.argument("name", default=None)
@click.option("-p", "--patchspec", default=None)
@connection_options
@common_options
def edit_stage(
    name,
    patchspec,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    if not name:
        click.secho("No stage specified. Pick one of the stages below:  ", fg="red")
        click_print_applied_stages(controller)
        return

    offset, scopedname = name.rsplit("/", 1)
    rule = controller.adageobj.view(offset).getRule(scopedname)

    s = yaml.safe_dump(rule.rule.stagespec, default_flow_style=False)

    if patchspec:
        edited = yaml.safe_load(open(patchspec))
    else:
        edited = yaml.safe_load(click.edit(s, editor="vi"))
    controller.patch_rule(rule.identifier, edited)
    click.secho("updated {}".format(name), fg="green")


@mancli.command()
@click.argument("name", default=None, nargs=-1)
@connection_options
@common_options
def undo_stage(
    name,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    if not name:
        click.secho("No stage specified. Pick one of the stages below:  ", fg="red")
        click_print_applied_stages(controller)
        return

    for n in name:
        offset, scopedname = n.rsplit("/", 1)
        rule = controller.adageobj.view(offset).getRule(scopedname)
        controller.undo_rules([rule.identifier])
        click.secho("undone {}".format(n), fg="green")


@mancli.command()
@click.argument("name", default=None)
@connection_options
@common_options
def remove_stage(
    name,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    if not name:
        click.secho("No stage specified. Pick one of the stages below:  ", fg="red")
        click_print_applied_stages(controller)
        return

    offset, name = name.rsplit("/", 1)
    rule = controller.adageobj.view(offset).getRule(name)
    controller.remove_rules([rule.identifier])


@mancli.command()
@click.argument("name")
@connection_options
@common_options
def reset_stage(
    name,
    metadir,
    accept_metadir,
    controller,
    ctrlopt,
    modelsetup,
    modelopt,
    backend,
    local,
    verbosity,
):
    handle_common_options(verbosity)
    ys = handle_connection_options(
        metadir,
        accept_metadir,
        controller,
        ctrlopt,
        modelsetup,
        modelopt,
        backend,
        local,
    )
    controller = ys.controller

    offset, name = name.rsplit("/", 1)
    rule = controller.adageobj.view(offset).getRule(name)
    if not rule:
        click.secho("state not found!", fg="red")

    r2s, _, _ = utils.rule_steps_indices(controller.adageobj)
    steps_of_rule = r2s[rule.identifier]
    controller.reset_nodes(steps_of_rule)


if __name__ == "__main__":
    mancli()



import logging

import yadage.backends.caching as caching
import yadage.backends.federatedbackend as federatedbackend
from packtivity.asyncbackends import ForegroundBackend
from packtivity.backendutils import backend_from_string

log = logging.getLogger(__name__)


class PacktivityBackend(federatedbackend.FederatedBackend):
    """
    a backend that mainly submits step tasks to a packtivity backend
    except for init nodes, which are resolvable trivially.
    """

    def __init__(
        self, packtivity_backendstring=None, packtivity_backend=None, backendopts=None
    ):
        purepubopts = backendopts.pop("purepubopts", {})
        if packtivity_backendstring:
            is_sync, backend = backend_from_string(
                packtivity_backendstring, backendopts=backendopts
            )
            assert not is_sync
        elif packtivity_backend:
            backend = packtivity_backend
        else:
            raise RuntimeError("need backend or backendstring")
        self.cached = False
        super(PacktivityBackend, self).__init__(
            {"purepub": ForegroundBackend(purepubopts), "packtivity": backend}
        )

    def enable_cache(self, cachestring=None, cache=None):
        if not (cache or cachestring):
            raise RuntimeError(
                "need to provide either cache object or cache config string"
            )

        self.cached = True
        self.backends["packtivity"] = caching.CachedBackend(
            self.backends["packtivity"],
            cache=cache if cache else caching.setupcache_fromstring(cachestring),
        )

    def routedbatchsubmit(self, tasks):
        actual_submits = [
            not t.metadata["wflow_hints"].get("is_purepub", False) for t in tasks
        ]
        if all(actual_submits) and not self.cached:
            try:
                submit_data = [
                    (task.spec, task.parameters, task.state, task.metadata)
                    for task in tasks
                ]
                return self.backends["packtivity"].batch_submit(*zip(*submit_data))
                log.warning(actual_submits, self.backends["packtivity"].batched_submit)
            except AttributeError:
                pass
        raise NotImplementedError

    def routedsubmit(self, task):
        is_pure_publishing = task.metadata["wflow_hints"].get("is_purepub", False)
        if is_pure_publishing:
            foreground_proxy = self.backends["purepub"].submit(
                task.spec, task.parameters, task.state, task.metadata
            )
            foreground_proxy.set_details({"labels": {"backend_hints": "is_purepub"}})
            return foreground_proxy

        else:
            # this is a little hacky, because the packtivity backends
            # take unrolled spec/parameters/context while the adage API
            # takes generalized task objects
            # possibly could use Munch on the packtivity side to
            # dynammicaly create .task/.parameters/.state-able objects
            if self.cached:
                # Cached backends adhere to the task-based API
                return self.backends["packtivity"].submit(task)
            else:
                # primary packtivity backends adhere to the unrolled API
                return self.backends["packtivity"].submit(
                    task.spec, task.parameters, task.state, task.metadata
                )

    def expected_result(self, task):
        if self.cached:
            # Cached backends adhere to the task-based API
            return self.backends["packtivity"].expected_result(task)
        else:
            # primary packtivity backends adhere to the unrolled API
            return self.backends["packtivity"].prepublish(
                task.spec, task.parameters, task.state
            )

    def routeproxy(self, proxy):
        details = proxy.details() or {}
        if details.get("labels", {}).get("backend_hints") == "is_purepub":
            return "purepub", proxy
        else:
            return "packtivity", proxy
        raise NotImplementedError("needs implementation")



import json
import logging
import shutil
import tempfile
import click
import yaml
import os
import base64
import glob

from .handlers.expression_handlers import handlers as exh
from .utils import process_refs
from .wflow import YadageWorkflow


def printRef(ref, dag, indent=""):
    click.secho(
        "{}name: {} position: {}, value: {}, id: {}".format(
            indent,
            dag.getNode(ref.stepid).name,
            ref.pointer.path,
            ref.pointer.resolve(dag.getNode(ref.stepid).result),
            ref.stepid,
        ),
        fg="cyan",
    )


def wflow_with_trivial_backend(instance):

    stateopts = {}
    wflow = YadageWorkflow.fromJSON(json.load(open(instance)), stateopts)
    return wflow


@click.group()
def utilcli():
    pass


@utilcli.command()
@click.argument("instance")
@click.argument("selection")
@click.option("--viewscope", default="")
@click.option("-v", "--verbosity", default="INFO")
def testsel(instance, selection, verbosity, viewscope):

    logging.basicConfig(
        level=getattr(logging, verbosity),
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    wflow = wflow_with_trivial_backend(instance)

    selresult = exh["stage-output-selector"](
        wflow.view(viewscope), yaml.safe_load(selection)
    )

    if not selresult:
        click.secho("Bad selection {}".format(selresult), fg="red")
        return

    click.secho(
        json.dumps(
            process_refs(selresult, wflow.dag),
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
        ),
        fg="green",
    )


@utilcli.command()
@click.argument("instance")
@click.argument("vizpdf")
@click.option("--viewscope", default="")
@click.option("-v", "--verbosity", default="INFO")
def viz(instance, vizpdf, viewscope, verbosity):
    logging.basicConfig(
        level=getattr(logging, verbosity),
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    import yadage.visualize as visualize

    wflow = wflow_with_trivial_backend(instance)

    dirpath = tempfile.mkdtemp()
    visualize.write_prov_graph(dirpath, wflow, scope=viewscope)
    shutil.copy("{}/yadage_workflow_instance.pdf".format(dirpath), vizpdf)
    shutil.rmtree(dirpath)


@utilcli.group()
def k8s():
    pass


@k8s.command()
@click.option("--hostname", default="docker-for-desktop")
@click.option("--path", default=None)
def create_state(hostname, path):
    pvc_name = "yadagedata"
    sc_name = "local-storage"
    path_base = path or os.getcwd()
    size = "1G"
    kubeyaml = """\
kind: PersistentVolumeClaim
apiVersion: v1
metadata:
  name: {pvc_name}
spec:
  accessModes:
    - ReadWriteMany
  resources:
    requests:
      storage: {size}
  storageClassName: {sc_name}
---
kind: StorageClass
apiVersion: storage.k8s.io/v1
metadata:
  name: {sc_name}
provisioner: kubernetes.io/no-provisioner
volumeBindingMode: WaitForFirstConsumer
---
apiVersion: v1
kind: PersistentVolume
metadata:
  name: yadage-pv
spec:
  capacity:
    storage: {size}
  volumeMode: Filesystem
  accessModes:
  - ReadWriteMany
  persistentVolumeReclaimPolicy: Delete
  storageClassName: {sc_name}
  local:
    path: {path_base}
  nodeAffinity:
    required:
      nodeSelectorTerms:
      - matchExpressions:
        - key: kubernetes.io/hostname
          operator: In
          values:
          - {hostname}
""".format(
        pvc_name=pvc_name,
        sc_name=sc_name,
        base_path=path_base,
        size=size,
        path_base=path_base,
        hostname=hostname,
    )
    click.echo(kubeyaml)


@k8s.command()
@click.option("--path")
def create_secrets(path):
    files = glob.glob("{}/*".format(path or os.environ["PACKTIVITY_AUTH_LOCATION"]))
    data = {
        os.path.basename(f): base64.b64encode(open(f, "rb").read()).decode("ascii")
        for f in files
    }
    secrets = """\
apiVersion: v1
data: {}
kind: Secret
metadata:
  name: hepauth
  namespace: default
type: Opaque
""".format(
        json.dumps(data)
    )
    click.echo(secrets)


if __name__ == "__main__":
    utilcli()



import importlib
import logging
import os
from contextlib import contextmanager

import yadageschemas

from .steering_object import YadageSteering
from .strategies import get_strategy

log = logging.getLogger(__name__)


def run_workflow(*args, **kwargs):
    """
    convenience function around steering context, when no additional settings
    are desired.
    """
    with steering_ctx(*args, **kwargs):
        pass


def execute_steering(
    steering_object,
    updateinterval=0.02,
    loginterval=30,
    default_trackers=True,
    strategy=None,
    strategyopts=None,
    backend=None,
    cache=None,
):

    ys = steering_object
    ys.adage_argument(
        default_trackers=default_trackers,
        trackevery=loginterval,
        update_interval=updateinterval,
        recursive_updates=True,
    )

    if cache:
        if cache == "checksums":
            backend.enable_cache(
                ":".join([cache, os.path.join(ys.metadir, "cache.json")])
            )
        else:
            backend.enable_cache(cache)

    custom_tracker = os.environ.get("YADAGE_CUSTOM_TRACKER", None)
    if custom_tracker:
        modulename, trackerclassname = custom_tracker.split(":")
        module = importlib.import_module(modulename)
        trackerclass = getattr(module, trackerclassname)
        ys.adage_argument(additional_trackers=[trackerclass()])

    if strategy is not None:
        ys.adage_argument(**get_strategy(strategy, strategyopts))

    ys.run_adage(backend)


@contextmanager
def steering_ctx(
    dataarg,
    workflow=None,
    initdata=None,
    toplevel=os.getcwd(),
    backend=None,
    controller="frommodel",
    ctrlopts=None,
    workflow_json=None,
    cache=None,
    dataopts=None,
    updateinterval=0.02,
    loginterval=30,
    schemadir=yadageschemas.schemadir,
    metadir=None,
    strategy=None,
    strategyopts=None,
    validate=True,
    visualize=True,
    wflowopts=None,
    accept_metadir=False,
    modelsetup="inmem",
    modelopts=None,
):

    ys = YadageSteering.create(
        metadir=metadir,
        accept_metadir=True if (accept_metadir or cache) else False,
        dataarg=dataarg,
        dataopts=dataopts,
        wflowopts=wflowopts,
        workflow_json=workflow_json,
        workflow=workflow,
        toplevel=toplevel,
        schemadir=schemadir,
        validate=validate,
        initdata=initdata,
        modelsetup=modelsetup,
        modelopts=modelopts,
        controller=controller,
        ctrlopts=ctrlopts,
    )

    yield ys

    try:
        execute_steering(
            steering_object=ys,
            updateinterval=updateinterval,
            loginterval=loginterval,
            default_trackers=visualize,
            strategy=strategy,
            strategyopts=strategyopts,
            backend=backend,
            cache=cache,
        )
    finally:
        log.info("done. dumping workflow to disk.")
        ys.serialize()
    if visualize:
        log.info("visualizing workflow.")
        ys.visualize()



import logging

from packtivity import datamodel
from packtivity.statecontexts import load_state

from .utils import outputReference

log = logging.getLogger(__name__)


class packtivity_task(object):
    """
    packtivity task
    """

    def __init__(self, name, spec, state, parameters=None, inputs=None):
        self.metadata = {"name": name}
        self.inputs = inputs or []
        self.parameters = datamodel.create(
            parameters or {}, state.datamodel if state else None
        )
        self.spec = spec
        self.state = state

    def pubOnlyTask(self):
        return (self.spec["environment"] is None) and (self.spec["process"] is None)

    # (de-)serialization
    @classmethod
    def fromJSON(cls, data, deserialization_opts=None):
        instance = cls(
            data["metadata"]["name"],
            data["spec"],
            load_state(data["state"], deserialization_opts) if data["state"] else None,
            data["parameters"],
            inputs=list(map(outputReference.fromJSON, data["inputs"])),
        )
        instance.metadata.update(**data["metadata"])
        return instance

    def json(self):
        serialized = {
            "metadata": self.metadata,
            "parameters": self.parameters.json(),
            "inputs": [x.json() for x in self.inputs],
            "type": "packtivity_task",
            "spec": self.spec,
            "state": self.state.json() if self.state else None,
        }
        return serialized



import json
import logging

from .wflow import YadageWorkflow
from .handlers.utils import handler_decorator

log = logging.getLogger(__name__)

modelhandlers, statemodel = handler_decorator()


@statemodel("inmem")
def inmem_model(modelsetup, modelopts, initmodel):
    return initmodel


@statemodel("filebacked")
def filebacked_model(modelsetup, modelopts, initmodel):
    filename = modelsetup.split(":")[-1]
    model = FileBackedModel(
        filename=filename, initmodel=initmodel, deserialization_opts=modelopts
    )
    return model


@statemodel("mongo")
def mongo_model(modelsetup, modelopts, initmodel):
    model = MongoBackedModel(initmodel=initmodel, deserialization_opts=modelopts)
    return model


def load_model_fromstring(modelsetup, modelopts=None, initmodel=None):
    modelopts = modelopts or {}
    for k in modelhandlers.keys():
        if modelsetup.startswith(k):
            return modelhandlers[k](modelsetup, modelopts, initmodel)
    raise RuntimeError("unknown state model %s" % modelsetup)


class MongoBackedModel(object):
    """
    model that holds the workflow state in a MongoDB database.
    """

    def __init__(
        self,
        deserialization_opts=None,
        connect_string="mongodb://localhost:27017/",
        initmodel=None,
        wflowid=None,
    ):
        from pymongo import MongoClient
        from bson.objectid import ObjectId

        self.deserialization_opts = deserialization_opts
        self.client = MongoClient(connect_string)
        self.db = self.client.wflowdb
        self.collection = self.db.workflows
        if initmodel:
            insertion = self.collection.insert_one(initmodel.json())
            self.wflowid = insertion.inserted_id
            log.info("created new workflow object with id %s", str(self.wflowid))
        if wflowid:
            self.wflowid = ObjectId(wflowid)

    def commit(self, data):
        """
        :param data: data to commit to disk. needs to have '.json()' method
        return: None
        """
        self.collection.replace_one({"_id": self.wflowid}, data.json())

    def load(self):
        """
        param load: load data from database
        return: yadage workflow object
        """
        jsondata = self.collection.find_one({"_id": self.wflowid})
        return YadageWorkflow.fromJSON(jsondata, self.deserialization_opts)


class FileBackedModel(object):
    """
    model that holds data on disk in a JSON file
    """

    def __init__(self, filename, deserialization_opts=None, initmodel=None):
        self.filename = filename
        self.deserialization_opts = deserialization_opts
        if initmodel:
            self.commit(initmodel)

    def commit(self, data):
        """
        :param data: data to commit to disk. needs to have '.json()' method
        """
        log.debug("committing model")
        jsondata = data.json()

        with open(self.filename, "w") as statefile:
            json.dump(jsondata, statefile)

    def load(self):
        """
        :return: the adage workflow object holding rules and the graph
        """
        log.debug("loading model")
        with open(self.filename) as statefile:
            jsondata = json.load(statefile)
            return YadageWorkflow.fromJSON(jsondata, self.deserialization_opts)



import logging
import jsonpath_rw

from .utils import handler_decorator
from ..utils import pointerize

log = logging.getLogger(__name__)

handlers, expression = handler_decorator()


def select_from_scope(scope, output):
    view = scope["view"]
    value = view.getValue(output)
    if not value:
        return None
    log.debug("resolved to %s", value)
    if isinstance(value, dict) and "expression_type" in value:
        log.debug("looking up expression %s", value)
        return handlers[value["expression_type"]](view, value)
    else:
        raise RuntimeError("not an expression value")


def select_reference(step, selection, is_scopes):
    """
    resolves a jsonpath selection on a step's result data and returns JSONPointerized matches

    :param step: the step holding the result data
    :param selection: a JSONPath expression

    :return: the first (and single) match of the expression as a JSON pointer
    """
    if is_scopes:
        return select_from_scope(step, selection)

    selection = selection or "$"
    log.debug("selecting %s from step %s", selection, step)
    pointerized = pointerize(step["result"], asref=True, stepid=step["id"])

    try:
        return pointerized[selection]
    except:
        pass

    matches = jsonpath_rw.parse(selection).find(pointerized)
    log.info("matches")
    if not matches:
        log.error(
            "no matches found for selection %s in result %s"
            % (selection, step["result"])
        )
        raise RuntimeError(
            "no matches found in reference selection. selection %s | result %s"
            % (selection, step["result"])
        )

    if len(matches) > 1:
        log.error(
            "found multiple matches to query: %s within result: %s\n \ matches %s",
            selection,
            step["result"],
            matches,
        )
        raise RuntimeError("multiple matches in result jsonpath query")
    return matches[0].value


def combine_outputs(outputs, flatten, unwrapsingle):
    """
    combines the result of multiple reference selections into a single outputs.
    non-list values will just be combined into a list while lists will be concatenated
    optinally we can return the sole element of a single-value list (argument: unwrapsingle)

    :param outputs:

    """
    log.debug("flatten %s unwrap: %s", flatten, unwrapsingle)
    combined = []
    for reference in outputs:
        if type(reference) == list:
            if flatten:
                for elementref in reference:
                    combined += [elementref]
            else:
                combined += [reference]
        else:
            combined += [reference]
    if len(combined) == 1 and unwrapsingle:
        combined = combined[0]
    return combined


def select_steps(stageview, query):
    """
    selects the step objects from the stage view based on a query and converts them to simple
    dictionaries with id,result keys

    :param stageview: the view object on which to perform the query
    :param query: a slection query (JSONPath expression)

    :return: list of {id: xx, result: yy} dictionaries
    """
    possible_steps = stageview.getSteps(query)
    possible_scopes = stageview.getScopes(query)
    if possible_steps and possible_scopes:
        raise RuntimeError("both steps and scopes in selection? should be impossible")
    if possible_scopes:
        return "scopes", [{"view": stageview.view(s)} for s in possible_scopes]
    if possible_steps:
        return (
            "steps",
            [
                {"id": s.identifier, "result": s.result}
                for s in stageview.getSteps(query)
            ],
        )
    return None, []


def select_outputs(steps, selection, flatten, unwrapsingle, is_scopes=False):
    return combine_outputs(
        map(lambda s: select_reference(s, selection, is_scopes), steps),
        flatten,
        unwrapsingle,
    )


@expression("stage-output-selector")
def stage_output_selector(stageview, selection):
    """
    :param stageview: the workflow view objct
    :param selection: the JSON-like selection dictionary
    :return :
    resolves a output reference by selecting the stage and stage outputs
    """
    log.debug("resolving selection %s", selection)
    if type(selection) is not dict:
        return None
    else:
        if "stages" in selection or "steps" in selection:
            if "stages" in selection and "steps" in selection:
                raise RuntimeError("stages and steps are aliases. pick one.")
            steps_selector = selection.get("stages") or selection.get("steps")
            selection_type, steps = select_steps(stageview, steps_selector)
            log.debug("selected steps %s %s", len(steps), steps)
            outputs = select_outputs(
                steps,
                selection.get("output"),
                selection.get("flatten", False),
                selection.get("unwrap", False),
                is_scopes=(selection_type == "scopes"),
            )
            log.debug("selected outputs %s", outputs)
            return outputs
        elif "step" in selection:
            selection_type, steps = select_steps(stageview, selection["step"])
            assert len(steps) == 1
            step = steps[0]
            return select_reference(
                step, selection.get("output"), selection_type == "scopes"
            )
        raise RuntimeError("not sure how to deal with this.")


@expression("fromvalue")
def value_resolver(view, expression):
    if "scope" in expression:
        scopes = view.getScopes(expression["scope"])
        assert len(scopes) == 1
        scope = scopes[0]
    else:
        scope = ""

    view = view.view(scope)

    value = select_from_scope({"view": view}, expression["key"])
    if not value:
        return None
    log.debug("resolved to %s", value)
    if isinstance(value, dict) and "expression_type" in value:
        log.debug("looking up expression %s", value)
        return handlers[value["expression_type"]](view, value)
    else:
        log.info("not an expression value %s", value)
        return value



from packtivity.backendutils import proxyloader, load_proxy
from yadage.backends.trivialbackend import TrivialProxy


class CachedProxy(object):
    def __init__(self, proxy, cacheid):
        self.proxy = proxy
        self.cacheid = cacheid

    def details(self):
        return {}

    def json(self):
        return {
            "proxyname": "CachedProxy",
            "proxy": self.proxy.json(),
            "cacheid": self.cacheid,
        }

    @classmethod
    def fromJSON(cls, data, deserialization_opts=None):
        deserialization_opts = deserialization_opts or {}
        p = load_proxy(data["proxy"], deserialization_opts, best_effort_backend=False)
        return cls(p, data["cacheid"])


@proxyloader("CachedProxy")
def cache_loader(jsondata, deserialization_opts=None, best_effort_backend=False):
    proxy = CachedProxy.fromJSON(jsondata, deserialization_opts)
    if best_effort_backend:
        raise NotImplementedError("nope")
    return proxy


@proxyloader("TrivialProxy")
def trivial_loader(jsondata, deserialization_opts=None, best_effort_backend=False):
    proxy = TrivialProxy.fromJSON(jsondata)
    if best_effort_backend:
        raise NotImplementedError("nope")
    return proxy



import copy
import itertools
import logging

import jq
import jsonpointer

import yadage.handlers.utils as utils

from ..stages import JsonStage
from ..tasks import packtivity_task
from ..utils import (
    init_stage_spec,
    leaf_iterator_jsonlike,
    outputReference,
    process_jsonlike,
    pointerize,
)
from .expression_handlers import handlers as exprhandlers

log = logging.getLogger(__name__)

handlers, scheduler = utils.handler_decorator()

# A scheduler does the following things:
#   - attached new nodes to the DAG
# - for each added step
#     - the step is given a name
#     - the step parameters are determined using the scheduler spec and context
#     - a list of used inputs (in the form of [stepname,outputkey,index])


def isExpression(parameter):
    return type(parameter) == dict and "expression_type" in parameter


def select_parameter(wflowview, parameter):
    """
    Evaluates parameter expressions (if needed) in the context of a workflow view

    :param wflowview: the workflow view on which to evaluete possible value expressions
    :param parameter: either a non-dict value or a JSON-like dict for a
                      supported value expression
    :return: the parameter value
    """
    log.debug("selecting parameter %s", parameter)
    if isExpression(parameter):
        handler = exprhandlers[parameter["expression_type"]]
        value = handler(wflowview, parameter)
    else:
        value = parameter
    return value


def finalize_value(wflowview, value, inputs):
    """
    finalize a value by recursively resolving references

    :param wflowview: the workflow view against which to resolve upstream references
    :param step: the step for which to track usage of upstream references
    :param value: the parameter value. May be a output reference, or a JSON value type
    :return: finalized parameter value
    """
    if type(value) == outputReference:
        return wflowview.dag.getNode(value.stepid).readfromresult(
            value.pointer.path, inputs
        )
    else:
        return value


def finalize_input(jsondata, wflowview):
    """
    evaluate final values of step parameters by resolving
    references to a upstream output and contextualizing stateful
    parameters. Also tracks usage of upstream references for the step

    :param wflowview: the workflow view view against which to resolve any upstream references
    :param step: the step that for which to track usage of upstream references
    :param jsondata: the prospective step parameters

    :return: finalized step parameters
    """
    result = copy.deepcopy(jsondata)
    inputs = []
    for leaf_pointer, leaf_value in leaf_iterator_jsonlike(jsondata):
        v = finalize_value(wflowview, leaf_value, inputs)
        if leaf_pointer.path == "":
            return v, inputs  # only one leaf
        leaf_pointer.set(result, v)
    return result, inputs


def step_or_stages(name, spec, inputs, parameters, state_provider, stageview):
    """
    :param name: name of the eventual (init-)step
    :param spec: the stage spec
    :param state_provider: the stage's state provider
    :param dependencies: dependent nodes

    :return: yadage or init step object
    """

    dependencies = [stageview.dag.getNode(k.stepid) for k in inputs]
    depstates = [d.task.state for d in set(dependencies) if d.task.state]

    if "step" in spec:
        step_state = state_provider.new_state(name, depstates)
        p = packtivity_task(
            name=name,
            spec=spec["step"],
            parameters=parameters,
            state=step_state,
            inputs=inputs,
        )
        return p, None
    elif "workflow" in spec:
        opts = spec.get("workflow_opts", {})

        init_spec = init_stage_spec(
            parameters.json(),
            discover=opts.get("discover", False),
            relative=opts.get("relative", True),
            used_inputs=[x.json() for x in inputs],
            name="init",
            nodename="init_{}".format(name),
        )
        stages = [init_spec] + spec["workflow"]["stages"]
        new_provider = state_provider.new_provider(name, init_states=depstates)
        stageobjects = [JsonStage(s, new_provider) for s in stages]
        return None, stageobjects
    elif "cases" in spec:
        for x in spec["cases"]:
            if parameters.jq(x["if"]).json():
                log.info("selected case %s", x["if"])
                return step_or_stages(
                    name, x, inputs, parameters, state_provider, stageview
                )
        log.info("no case selected on pars %s", parameters)
        return None, None
    raise RuntimeError(
        "do not know what kind of stage spec we are dealing with. %s" % spec.keys()
    )


def registerExpressions(stage, expressions):
    if not expressions:
        return
    for key, expression in expressions.items():
        stage.view.addValue(key, expression)


def addStepOrWorkflow(name, stage, parameters, inputs, spec):
    """
    adds a step or a sub-workflow based on a init step

    :param str name: the name of the step or sub-workflow
    :param stage: the stage from which to use state context and workflow view
    :param step: either a packtivity_task (for normal workflow steps) initstep object (for sub-workflows)
    :param spec: the stage spec

    :return: None
    """
    step, stages = step_or_stages(
        name, spec, inputs, parameters, stage.state_provider, stage.view
    )
    if step:
        stage.addStep(step)
        log.debug("scheduled a step")

    if stages:  # subworkflow case
        stage.addWorkflow(stages, isolate=True)
        log.debug("scheduled a subworkflow")


def get_parameters(parameters):
    """
    retrieve parameters from the spec

    :param spec: the stage spec
    :return: a JSON-like object of stage parameters
    """
    return {x["key"]: x["value"] for x in parameters}


@scheduler("singlestep-stage")
def singlestep_stage(stage, spec):
    """
    a simple state that adds a single step/workflow. The node is attached
    to the DAG based on used upstream outputs

    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    """
    log.debug("scheduling singlestep stage with spec:\n%s", spec)
    parameters = {
        k: select_parameter(stage.view, v)
        for k, v in get_parameters(spec["parameters"]).items()
    }
    finalized, inputs = finalize_input(parameters, stage.view)
    finalized = stage.datamodel.create(
        finalized, getattr(stage.state_provider, "datamodel", None)
    )
    addStepOrWorkflow(stage.name, stage, finalized, inputs, spec)
    registerExpressions(stage, spec.get("register_values"))


def chunk(alist, chunksize):
    """split a list into equal-sized chunks of size chunksize"""
    return [alist[x : x + chunksize] for x in range(0, len(alist), chunksize)]


def partition(alist, partitionsize):
    """split a list into partitionsize parts"""
    total_len = len(alist)
    if partitionsize > total_len:
        partitionsize = total_len
    assert partitionsize <= total_len
    end = 0
    partitioned = []
    for k in range(partitionsize):
        begin = end
        end = end + (total_len + k) // partitionsize
        partitioned.append(alist[begin:end])
    return partitioned


def groupmany(iterable, batchsize=None, partitionsize=None):
    if batchsize:
        return chunk(list(iterable), batchsize)
    if partitionsize:
        return partition(list(iterable), partitionsize)
    return iterable


def scatter(parameters, scatter, batchsize=None, partitionsize=None):
    """
    convert a parameter set and scatter definition into a list
    of single parameter sets.

    :param parameters: the parameter definition
    :param scatter: scattering method. One of 'zip' or 'cartesian'

    :return: list of parameter sets
    """

    log.debug(
        "scattering:  pars: %s scatter: %s batchsisze: %s partitionsize: %s",
        parameters,
        scatter,
        batchsize,
        partitionsize,
    )
    commonpars = parameters.copy()
    to_scatter = {}
    for scatpar in scatter["parameters"]:
        to_scatter[scatpar] = groupmany(
            commonpars.pop(scatpar), batchsize, partitionsize
        )

    singlesteppars = []
    if scatter["method"] == "zip":
        keys, zippable = zip(*[(k, v) for k, v in to_scatter.items()])
        for zipped in zip(*zippable):
            individualpars = dict(zip(keys, zipped))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]

    if scatter["method"] == "cartesian":
        for what in itertools.product(*[to_scatter[k] for k in scatter["parameters"]]):
            individualpars = dict(zip(scatter["parameters"], what))
            pars = commonpars.copy()
            pars.update(**individualpars)
            singlesteppars += [pars]
    return singlesteppars


@scheduler("multistep-stage")
def multistep_stage(stage, spec):
    """
    a stage that attaches an array of nodes to the DAG. The number of nodes
    is determined by a scattering recipe. Currently two algs are supported

    - ``zip``: one or more arrays of length n are iterated through in lock-step.
       n nodes are added to the DAG where the  parameters values are set to
           the values in the iteration
    - ``cartesian``:  a cartesian product of a number of arrays (possibly different sizes)
       adds n1 x n2 x ... nj nodes.

    Nodes are attached to the DAG based on used upstream inputs

    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    """
    log.info("scheduling multistep stage with spec:\n%s", spec)
    log.debug("selecting parameters")
    parameters = {
        k: select_parameter(stage.view, v)
        for k, v in get_parameters(spec["parameters"]).items()
    }
    log.info("scattering")
    singlesteppars = scatter(
        parameters, spec["scatter"], spec.get("batchsize"), spec.get("partitionsize")
    )

    log.info("scattering")

    for i, pars in enumerate(singlesteppars):
        singlename = "{}_{}".format(stage.name, i)
        finalized, inputs = finalize_input(pars, stage.view)
        finalized = stage.datamodel.create(
            finalized, getattr(stage.state_provider, "datamodel", None)
        )
        addStepOrWorkflow(singlename, stage, finalized, inputs, spec)
    registerExpressions(stage, spec.get("register_values"))


def process_noderef(leafobj, resultscript, view):
    n = view.dag.getNode(leafobj["_nodeid"])
    return jq.jq(resultscript).transform(
        pointerize(n.result, False, n.identifier), multiple_output=True
    )


def process_wflowref(leafobj, view):
    nodeselector, resultscript = leafobj["$wflowref"]
    nodestruct = jq.jq(nodeselector).transform(view.steps, multiple_output=True)
    return process_jsonlike(
        nodestruct, 'has("_nodeid")', lambda x: process_noderef(x, resultscript, view)
    )


def process_wflowpointer(leafobj):
    p = leafobj["$wflowpointer"]
    return outputReference(p["step"], jsonpointer.JsonPointer(p["result"]))


@scheduler("jq-stage")
def jq_stage(stage, spec):
    """
    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    """
    binds = spec["bindings"]
    binds = process_jsonlike(
        binds, 'has("$wflowref")', lambda x: process_wflowref(x, stage.view)
    )
    log.info("transforming binds: %s", binds)
    stagescript = spec["stepscript"]
    singlesteps = jq.jq(stagescript).transform(binds, multiple_output=False)

    singlesteppars = map(
        lambda x: process_jsonlike(x, 'has("$wflowpointer")', process_wflowpointer),
        singlesteps,
    )

    postscript = spec["postscript"]
    for i, pars in enumerate(singlesteppars):
        singlename = "{}_{}".format(stage.name, i)

        finalized, inputs = finalize_input(pars, stage.view)
        log.info("postscripting: %s", finalized)
        after_post = jq.jq(postscript).transform(finalized, multiple_output=False)
        after_post = stage.datamodel.create(after_post)
        log.info("finalized to: %s", after_post)
        addStepOrWorkflow(singlename, stage, after_post, inputs, spec)
    registerExpressions(stage, spec.get("register_values"))


@scheduler("init-stage")
def init_stage(stage, spec):
    """
    :param stage: common stage parent object
    :param spec: stage JSON-like spec

    :return: None
    """
    inputs = []
    if spec.get("inputs"):
        inputs = list(map(outputReference.fromJSON, spec["inputs"]))
        log.debug("initializing scope from dependent tasks with inputs")
    else:
        log.debug("initializing scope from dependent tasks")

    depstates = stage.state_provider.init_states if stage.state_provider else []

    if stage.state_provider:
        step_state = stage.state_provider.new_state(
            stage.name, depstates, readonly=True
        )
    else:
        step_state = None

    init_spec = spec["step"]
    task = packtivity_task(
        spec["nodename"] or stage.name,
        init_spec,
        state=step_state,
        parameters=spec["parameters"],
        inputs=inputs,
    )
    stage.addStep(task)
    registerExpressions(stage, spec.get("register_values"))



import logging
import os
import shutil
from six import string_types
import zipfile
from packtivity.statecontexts import load_state
from packtivity.statecontexts.posixfs_context import LocalFSState
import fnmatch

log = logging.getLogger(__name__)


import requests

try:
    # For Python 3.0 and later
    from urllib.request import urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen


def download_file(url, local_filename):
    if not "http" in url:
        f = urlopen(url)
        with open(local_filename, "wb") as lf:
            lf.write(f.read())
    else:
        h = (
            {}
            if not "YADAGE_INIT_TOKEN" in os.environ
            else {"PRIVATE-TOKEN": os.environ["YADAGE_INIT_TOKEN"]}
        )
        r = requests.get(url, stream=True, headers=h)
        with open(local_filename, "wb") as f:
            shutil.copyfileobj(r.raw, f)


def prepare_workdir_from_archive(initdir, inputarchive, match=None):
    url = inputarchive
    os.makedirs(initdir)
    localzipfile = "{}/.yadage_inputarchive.zip".format(initdir)
    download_file(url, localzipfile)

    if not match:
        with zipfile.ZipFile(localzipfile) as zf:
            zf.extractall(path=initdir)
    else:
        assert match.endswith("/")
        with zipfile.ZipFile(localzipfile) as zf:
            filtered = fnmatch.filter(zf.namelist(), match)
            assert len(filtered) == 1
            basepath = filtered[0]
            for x in zf.namelist():
                if x != basepath and basepath in x:
                    target = x.replace(basepath, "")
                    if target.endswith("/"):
                        os.makedirs(os.path.join(initdir, target))
                    else:
                        with open(os.path.join(initdir, target), "wb") as f:
                            f.write(zf.read(x))
    os.remove(localzipfile)


def _merge_states(lhs, rhs):
    return LocalFSState(lhs.readwrite + rhs.readwrite, lhs.readonly + rhs.readonly)


class LocalFSProvider(object):
    def __init__(self, *base_states, **kwargs):
        base_states = list(base_states)
        self.nest = kwargs.get("nest", True)
        self.ensure = kwargs.get("ensure")
        self.init_states = kwargs.get("init_states")
        self.sub_inits = kwargs.get("sub_init_states", {})

        first = base_states.pop()
        assert first

        self.base = first

        while base_states:
            next_state = base_states.pop()
            if not next_state:
                continue
            self.base = _merge_states(self.base, next_state)

    def new_provider(self, name, init_states=None):
        new_base_ro = self.base.readwrite + self.base.readonly
        new_base_rw = [os.path.join(self.base.readwrite[0], name)]

        sub_init = self.sub_inits.get(name, {}).get("_data")
        sub_inits = [sub_init] if sub_init else []
        init_states = init_states or []

        newsinits = {
            k: v for k, v in self.sub_inits.get(name, {}).items() if k != "_data"
        }
        return LocalFSProvider(
            LocalFSState(new_base_rw, new_base_ro),
            nest=self.nest,
            ensure=self.ensure,
            init_states=init_states + sub_inits,
            sub_init_states=newsinits,
        )

    def new_state(self, name, dependencies, readonly=False):
        """
        creates a new context from an existing context.

        if subdir is True it declares a new read-write nested under the old
        context's read-write and adds all read-write and read-only locations
        of the old context as read-only. This is recommended as it makes rolling
        back changes to the global state made in this context easy.

        else the same readwrite/readonly configuration as the parent context is used
        """

        readwrites = []
        if not readonly:
            readwrites = (
                ["{}/{}".format(self.base.readwrite[0], name)]
                if self.nest
                else self.base.readwrite
            )

        log.debug("new state is: rw: %s", readwrites)
        state_identifier = name.replace("/", "_")  # replace in case name is nested path
        newstate = LocalFSState(
            readwrite=readwrites, identifier=state_identifier, dependencies=dependencies
        )

        if not readonly and self.ensure:
            newstate.ensure()

        return newstate

    def json(self):
        def subinit_json(subinits):
            def handle(k, v):
                if k == "_data":
                    return v.json() if v else None
                return {kk: handle(kk, vv) for kk, vv in v.items()}

            return {k: handle(k, v) for k, v in subinits.items()}

        d = {
            "state_provider_type": "localfs_provider",
            "base_state": self.base.json(),
            "init_states": [s.json() for s in self.init_states]
            if self.init_states
            else [],
            "nest": self.nest,
            "ensure": self.ensure,
            "sub_init_states": subinit_json(self.sub_inits),
        }
        return d

    @classmethod
    def fromJSON(cls, jsondata, deserialization_opts):
        def subinit_load(subinit_data):
            def handle(k, v):
                if k == "_data":
                    return load_state(v, deserialization_opts) if v else None
                return {kk: handle(kk, vv) for kk, vv in v.items()}

            return {k: handle(k, v) for k, v in subinit_data.items()}

        instance = cls(
            load_state(jsondata["base_state"], deserialization_opts),
            nest=jsondata["nest"],
            ensure=jsondata["ensure"],
            init_states=[
                load_state(x, deserialization_opts) for x in jsondata["init_states"]
            ],
            sub_init_states=subinit_load(jsondata["sub_init_states"]),
        )
        return instance


def setup_provider(dataarg, dataopts):
    workdir = dataarg.split(":", 2)[1]
    read = dataopts.get("read", None)
    nest = dataopts.get("nest", True)
    ensure = dataopts.get("ensure", True)
    overwrite = dataopts.get("overwrite", False)
    subinits = dataopts.get("subinits", {})
    pathbase = dataopts.get("pathbase")

    if overwrite and os.path.exists(workdir):
        log.info("work directory %s will be purged and overwritten", workdir)
        shutil.rmtree(workdir)

    init_states = []

    initdir = os.path.join(pathbase or workdir, dataopts.get("initdir", "init"))
    inputarchive = dataopts.get("inputarchive", None)
    if initdir:
        if inputarchive:
            prepare_workdir_from_archive(
                initdir, inputarchive, match=dataopts.get("archivematch")
            )
        init_state = LocalFSState(readonly=[initdir])
        init_states.append(init_state)

    def handle_init_spec(pathbase, spec):
        subinits = {}
        for k, v in spec.items():
            if isinstance(v, string_types):
                subinits[k] = {
                    "_data": LocalFSState(readonly=[os.path.join(pathbase, v)])
                }
            if isinstance(v, dict):
                subinits.setdefault(k, {})["_data"] = (
                    LocalFSState(readonly=[os.path.join(pathbase, v.pop("_data"))])
                    if "_data" in v
                    else None
                )
                for kk, vv in handle_init_spec(pathbase, v).items():
                    subinits.setdefault(k, {})[kk] = vv
        return subinits

    if subinits:
        subinits = handle_init_spec(pathbase or workdir, subinits)

    writable_state = LocalFSState([workdir])
    return LocalFSProvider(
        read,
        writable_state,
        ensure=ensure,
        nest=nest,
        init_states=init_states,
        sub_init_states=subinits,
    )



from packtivity.asyncbackends import PacktivityProxyBase
from packtivity import datamodel as _datamodel


class TrivialProxy(PacktivityProxyBase):
    """
    A trivial proxy that carries the results and status already
    """

    def __init__(self, status, resultdata, datamodel):
        self.status = status
        self.resultdata = resultdata
        self.datamodel = datamodel

    def proxyname(self):
        return "TrivialProxy"

    def details(self):
        return {
            "resultdata": self.resultdata,
            "datamodel": self.datamodel,
            "status": self.status,
        }

    @classmethod
    def fromJSON(cls, data):
        return cls(**data["proxydetails"])


class TrivialBackend(object):
    """
    A trivial backend that can only return proxy based information, no submission
    """

    def __init__(self):
        self.datamodel = _datamodel

    def submit(self, task):
        raise NotImplementedError("The trivial proxy is not made for submission")

    def result(self, resultproxy):
        return self.datamodel.create(resultproxy.resultdata, resultproxy.datamodel)

    def expected_result(self, resultproxy):
        return None

    def ready(self, resultproxy):
        # when we have a proxy it is by definition ready...
        return True

    def successful(self, resultproxy):
        return resultproxy.status == "SUCCESS"

    def fail_info(self, resultproxy):
        return None



import logging
from adage.pollingexec import yes_man

from .interactive import custom_decider, decide_step, interactive_deciders
from .utils import advance_coroutine
from .handlers.utils import handler_decorator

log = logging.getLogger(__name__)

handlers, strategy = handler_decorator()


def get_strategy(strategystring, strategyopts=None):
    strategyopts = strategyopts or {}
    for k in handlers.keys():
        if strategystring.startswith(k):
            return handlers[k](strategystring, strategyopts)

    raise RuntimeError("Unknown Strategy %s", strategystring, handlers.keys())


@strategy("interactive")
def interactive_strategy(name, opts):
    extend, submit = interactive_deciders()
    return dict(extend_decider=extend, submit_decider=submit)


@strategy("askforsubmit")
def askforsubmit(name, opts=None):
    idbased = opts.get("idbased", False)
    extend_decider = yes_man()
    advance_coroutine(extend_decider)

    submit_decider = custom_decider(decide_step, idbased=idbased)()
    advance_coroutine(submit_decider)

    return dict(extend_decider=extend_decider, submit_decider=submit_decider)


@strategy("target")
def target(name, opts=None):
    opts = opts or {}
    idbased = opts.get("idbased", False)
    _, targetname = name.split(":", 1)
    targetoffset, targetname = targetname.rsplit("/", 1)

    def isfinished(controller, idbased):
        dag = controller.adageobj.dag
        targetnode = dag.getNodeByName(
            targetname,
            nodefilter=lambda x: x.task.metadata["wflow_offset"] == targetoffset,
        )
        if targetnode:
            return (targetnode.ready(), targetnode.successful())
        return False, False

    def upstream(dag, target_identifier):
        predecessors = dag.predecessors(target_identifier)
        return list(
            set(predecessors + [g for p in predecessors for g in upstream(dag, p)])
        )

    def submit_if_node_is_upstream(node, controller, idbased):
        dag = controller.adageobj.dag
        if idbased:
            node = dag.getNode(node)
        matched_nodes = controller.adageobj.view(targetoffset).getSteps(targetname)
        assert len(matched_nodes) <= 1
        target_node = matched_nodes[0] if matched_nodes else None
        node_identifier = node.identifier
        log.debug(
            "decide if submitting node: %s %s %s", node, node.identifier, target_node
        )
        if target_node:
            target_identifier = target_node.identifier
            target_upstream = upstream(dag, target_identifier)
            if node_identifier == target_identifier:
                return True
            if not node_identifier in target_upstream:
                log.debug("not submitting %s since it is not in upstream of target")
                return False
        return True

    extend_decider = yes_man()
    advance_coroutine(extend_decider)

    submit_decider = custom_decider(submit_if_node_is_upstream, idbased=idbased)()
    advance_coroutine(submit_decider)

    finish_decider = custom_decider(isfinished, idbased=idbased, unroll_data=False)()
    advance_coroutine(finish_decider)

    return dict(
        extend_decider=extend_decider,
        submit_decider=submit_decider,
        finish_decider=finish_decider,
    )



import logging

from .steering_object import YadageSteering
from .utils import setupbackend_fromstring
from .wflow import YadageWorkflow
from .controllers import YadageController

log = logging.getLogger(__name__)


def connect(
    metadir, accept_metadir, ctrlstring, ctrlopts, modelsetup, modelopts, backendstring
):
    ys = YadageSteering.connect(
        accept_metadir=accept_metadir,
        metadir=metadir,
        ctrlstring=ctrlstring,
        ctrlopts=ctrlopts,
        modelsetup=modelsetup,
        modelopts=modelopts,
    )
    if backendstring:
        ys.controller.backend = setupbackend_fromstring(backendstring)
    return ys


def preview_rule(wflow, name=None, identifier=None):
    stateopts = {}
    newflow = YadageWorkflow.fromJSON(wflow.json(), stateopts)
    YadageController(newflow).sync_backend()
    if identifier:
        rule = newflow.view().getRule(identifier=identifier)
    else:
        offset, name = name.split("/")
        rule = newflow.view(offset).getRule(name)

    if not rule.applicable(newflow):
        log.warning("rule not applicable")
        return

    rule.apply(newflow)
    newflow.rules.remove(rule)
    newflow.applied_rules.append(rule)

    existing_rules = [x.identifier for x in (wflow.rules + wflow.applied_rules)]
    existing_nodes = wflow.dag.nodes()

    new_rules = [
        {"name": x.rule.name, "offset": x.offset}
        for x in newflow.rules
        if x.identifier not in existing_rules
    ]
    new_nodes = [
        {"name": newflow.dag.getNode(n).name, "parents": newflow.dag.predecessors(n)}
        for n in newflow.dag.nodes()
        if n not in existing_nodes
    ]
    return new_rules, new_nodes



import contextlib
import importlib
import logging
import os
from adage.wflowcontroller import BaseController
from packtivity.syncbackends import defaultsyncbackend

from .reset import collective_downstream, remove_rules, reset_steps, undo_rules
from .wflow import YadageWorkflow
from .handlers.utils import handler_decorator

log = logging.getLogger(__name__)

ctrlhandlers, controller = handler_decorator()


class YadageController(BaseController):
    def __init__(self, *args, **kwargs):
        self.prepublishing_backend = defaultsyncbackend()
        self.disable_backend = False
        self.disable_prepublishing = kwargs.pop("disable_prepub", False)
        super(YadageController, self).__init__(*args, **kwargs)

    def sync_expected(self):
        for n in self.adageobj.dag.nodes():
            if (
                "YADAGE_IGNORE_PREPUBLISHING" in os.environ
                or self.disable_prepublishing
            ):
                continue
            node = self.adageobj.dag.getNode(n)
            node.expected_result = self.prepublishing_backend.prepublish(
                node.task.spec, node.task.parameters.json(), node.task.state
            )

    def sync_backend(self):
        self.sync_expected()
        if not self.disable_backend:
            super(YadageController, self).sync_backend()


@controller("frommodel")
def frommodel_controller(ctrlstring, ctrlopts, model=None):
    if isinstance(model, YadageWorkflow):
        return YadageController(model, **ctrlopts)
    else:
        return PersistentController(model, **ctrlopts)


@controller("http")
def http_controller(ctrlstring, ctrlopts, model=None):
    try:
        from yadagehttpctrl.clientcontroller import YadageHTTPController

        ctrl = YadageHTTPController(server=ctrlstring, **ctrlopts)
        return ctrl
    except ImportError:
        log.exception("try installing yadagehttpctrl")


@controller("py:")
def frompython_controller(ctrlstring, ctrlopts, model=None):
    _, module, ctrlclass = ctrlstring.split(":")
    module = importlib.import_module(module)
    ctrlclass = getattr(module, ctrlclass)
    if ctrlopts.pop("pass_model", False):
        ctrlopts["model"] = model
    return ctrlclass(**ctrlopts)


def setup_controller(model=None, controller="frommodel", ctrlopts=None):
    """
    return controller instance based on state configuration. For
    transaction-based states, returns PersistentController, for in-
    memory states returns in BaseController
    """
    ctrlopts = ctrlopts or {}
    for k in ctrlhandlers.keys():
        if controller.startswith(k):
            return ctrlhandlers[k](controller, ctrlopts, model)
    raise RuntimeError("unknown controller type %s" % controller)


class PersistentController(YadageController):
    """
    workflow controller, that explicltly calls transaction methods on non read-only operations on the workflow state
    """

    def __init__(self, model, backend=None):
        """
        :param model: the model on whih the controller will operate
        :param backend: the backend to against which to check workflow state.

        :return: the controller instance
        """
        self.model = model
        super(PersistentController, self).__init__(self.model.load(), backend)

    @contextlib.contextmanager
    def transaction(self, sync=True):
        """the transaction context. will commit model to persistent store on exit."""
        self.adageobj = self.model.load()
        if sync:
            log.debug("syncing to setup tx %s", self)
            super(PersistentController, self).sync_backend()
        yield

        isvalid = self.validate()
        if not isvalid:
            # raise RuntimeError('was about to commit invalid data!')
            log.warning("commit is invalid %s", isvalid)
        if sync:
            log.debug("syncing to teardown tx %s", self)
            super(PersistentController, self).sync_backend()
        self.model.commit(self.adageobj)

    def submit_nodes(self, nodeids):
        """
        submit nodes to backend

        :param nodeids: list of ids of nodes to be submitted
        :return: None
        """
        log.debug("transaction to submit")
        with self.transaction():
            nodes = [self.adageobj.dag.getNode(nodeid) for nodeid in nodeids]
            # log.info('submitting nodes to backend: %s', nodes)
            super(PersistentController, self).submit_nodes(nodes)

    def apply_rules(self, ruleids):
        """
        apply rules to workflow instance

        :param ruleids: list of ids of reuls to be applied
        :return: None
        """
        log.debug("transaction to apply")
        with self.transaction():
            rules = [r for r in self.adageobj.rules if r.identifier in ruleids]
            super(PersistentController, self).apply_rules(rules)

    def sync_backend(self):
        """
        synchronize node data with backend
        """
        log.debug("transaction to sync but (without sync in tx)")
        with self.transaction(sync=False):  # disable sync to avoid infinite recursion
            super(PersistentController, self).sync_backend()

    def applicable_rules(self):
        """
        :return: a list of rules whose predicate is fulfilled
        """
        applicable_rules = super(PersistentController, self).applicable_rules()
        return [x.identifier for x in applicable_rules]

    def submittable_nodes(self):
        """

        :return: a list of nodes with sucessfull and completed upstream
        """
        submittable_nodes = super(PersistentController, self).submittable_nodes()
        return [x.identifier for x in submittable_nodes]

    def add_rules(self, rulespecs, dataarg, offset="", groupname=None, dataopts=None):
        log.debug("adding %s rules", len(rulespecs))
        from .stages import JsonStage
        from .state_providers import state_provider_from_string

        sp = state_provider_from_string(dataarg, dataopts)
        rules = [JsonStage(json, sp) for json in rulespecs]
        with self.transaction():
            self.adageobj.view(offset).addWorkflow(rules, groupname)

    def patch_rule(self, ruleid, patchspec):
        with self.transaction():
            rule = [x for x in self.adageobj.rules if x.identifier == ruleid][0]
            rule.rule.stagespec = patchspec

    def undo_rules(self, ruleids):
        """
        :param nodeids: list of ids of nodes to reset
        :return: None
        """
        with self.transaction():
            undo_rules(self.adageobj, ruleids)

    def remove_rules(self, ruleids):
        with self.transaction():
            remove_rules(self.adageobj, ruleids)

    def reset_nodes(self, nodeids):
        """
        :param nodeids: list of ids of nodes to reset
        :return: None
        """
        to_reset = nodeids + collective_downstream(self.adageobj, nodeids)
        with self.transaction():
            reset_steps(self.adageobj, to_reset)



import adage
from adage.serialize import obj_to_json

from .stages import JsonStage, OffsetStage
from .wflowview import WorkflowView
from .wflownode import YadageNode


class YadageWorkflow(adage.adageobject):
    """
    The overall workflow state object that extends the basic
    Adage state object by two bookkeeping structures.
    """

    def __init__(
        self,
        dag=None,
        rules=None,
        applied_rules=None,
        bookkeeping=None,
        stepsbystage=None,
        values=None,
    ):
        super(YadageWorkflow, self).__init__(
            dag=dag, rules=rules, applied_rules=applied_rules
        )
        self.stepsbystage = stepsbystage or {}
        self.bookkeeping = bookkeeping or {}
        self.values = values or {}

    def view(self, offset=""):
        return WorkflowView(self, offset)

    def json(self):
        json_or_nil = lambda x: None if x is None else x.json()
        data = obj_to_json(self, json_or_nil, json_or_nil)

        data["bookkeeping"] = self.bookkeeping
        data["stepsbystage"] = self.stepsbystage
        data["values"] = self.values
        return data

    @classmethod
    def fromJSON(cls, data, deserialization_opts=None, backend=None):
        def node_deserializer(data):
            node = YadageNode.fromJSON(data, deserialization_opts)
            if backend:
                # node.backend = backend
                node.update_state(backend=backend)
            return node

        def rule_deserializer(data):
            return OffsetStage.fromJSON(data, deserialization_opts)

        dag = adage.serialize.dag_from_json(data["dag"], node_deserializer)

        instance = cls(
            dag=dag,
            rules=[rule_deserializer(x) for x in data["rules"]],
            applied_rules=[rule_deserializer(x) for x in data["applied"]],
            bookkeeping=data["bookkeeping"],
            stepsbystage=data["stepsbystage"],
            values=data["values"],
        )

        return instance

    @classmethod
    def createFromJSON(cls, jsondata, state_provider):
        instance = cls()
        rules = [
            JsonStage(stagedata, state_provider) for stagedata in jsondata["stages"]
        ]
        instance.view().addWorkflow(rules)
        return instance



import logging
import os
import copy

import adage
import shutil
from .serialize import snapshot
from .wflowstate import load_model_fromstring
from .controllers import setup_controller
from .utils import setupbackend_fromstring, prepare_meta
from .creators import handlers as creators

log = logging.getLogger(__name__)


class YadageSteering(object):
    """
    high level steering object to manage worklfow execution
    """

    def __init__(self, metadir, controller):
        self.metadir = metadir
        self.controller = controller
        self.adage_kwargs = dict(workdir=os.path.join(metadir, "adage"))

    @classmethod
    def connect(
        cls,
        metadir,
        ctrlstring,
        ctrlopts=None,
        modelsetup=None,
        modelopts=None,
        accept_metadir=False,
    ):
        prepare_meta(metadir, accept_metadir)

        model = None
        if modelsetup:
            model = load_model_fromstring(modelsetup, modelopts)

        ctrl = setup_controller(model=model, controller=ctrlstring, ctrlopts=ctrlopts)
        log.info("connected to model")
        return cls(metadir, ctrl)

    @classmethod
    def create(cls, **kwargs):
        dataarg = kwargs["dataarg"]
        dataopts = kwargs.get("dataopts") or {}
        is_local_data = len(dataarg.split(":", 1)) == 1
        if is_local_data:
            metadir = kwargs.get("metadir")
            metadir = metadir or "{}/_yadage/".format(dataarg)
            if dataopts.get("overwrite") and os.path.exists(metadir):
                shutil.rmtree(metadir)
        else:
            metadir = kwargs["metadir"]
        accept_metadir = kwargs.pop("accept_metadir", False)

        kw = copy.deepcopy(kwargs)
        kw["metadir"] = metadir
        prepare_meta(
            metadir, accept_metadir
        )  # meta must be here because data model might store stuff here
        ctrl = creators["local"](**kw)
        prepare_meta(metadir, accept=True)  # Hack in case creator deletes meta
        return cls(metadir, ctrl)

    @property
    def workflow(self):
        """
        :return: the workflow object (from the controller)
        """
        return self.controller.adageobj

    def adage_argument(self, **kwargs):
        """
        add keyword arguments for workflow execution (adage)

        :param kwargs: adage keyword arguments (see adage documentation for options)
        """
        self.adage_kwargs.update(**kwargs)

    def run_adage(self, backend="auto", **adage_kwargs):
        """
        execution workflow with adage based against given backend
        :param backend: backend to use for packtivity processing.
        """
        if backend == "auto":
            # respect if the controller already has a backend wired up
            self.controller.backend = (
                self.controller.backend or setupbackend_fromstring("multiproc:auto")
            )
            log.info("backend automatically set to %s", backend)
        elif backend:
            self.controller.backend = backend

        assert self.controller.backend
        self.adage_argument(**adage_kwargs)
        adage.rundag(controller=self.controller, **self.adage_kwargs)

    def serialize(self):
        """
        serialized workflow and backend states (stored in meta directory)
        """
        snapshot(self.workflow, "{}/yadage_snapshot_workflow.json".format(self.metadir))

    def visualize(self):
        """
        generate workflow visualization (stored in meta directory)
        """
        import yadage.visualize as visualize

        visualize.write_prov_graph(self.metadir, self.workflow, vizformat="png")
        visualize.write_prov_graph(self.metadir, self.workflow, vizformat="pdf")



import click

from .utils import advance_coroutine


def decide_rule(rule, controller, idbased):
    click.secho("we could extend DAG with rule", fg="blue")
    if idbased:
        rule = next(x for x in controller.adageobj.rules if x.identifier == rule)

    click.secho("rule: {}/{} ({})".format(rule.offset, rule.rule.name, rule.identifier))
    # IPython.embed()
    try:
        resp = raw_input(click.style("Shall we? (y/N) ", fg="blue"))
    except NameError:
        resp = input(click.style("Shall we? (y/N) ", fg="blue"))

    shall = resp.lower() == "y"
    if shall:
        click.secho("ok we will extend.", fg="green")
    else:
        click.secho("maybe another time... your response: {}".format(resp), fg="yellow")
    return shall


def decide_step(node, controller, idbased):

    if idbased:
        node = controller.adageobj.dag.getNode(idbased)

    click.echo("we could submit a DAG node {}".format(node))
    # IPython.embed()
    try:
        resp = raw_input(click.style("Shall we? (y/N) ", fg="magenta"))
    except NameError:
        resp = input(click.style("Shall we? (y/N) ", fg="magenta"))

    shall = resp.lower() == "y"
    if shall:
        click.secho("ok we will submit.", fg="green")
    else:
        click.secho(
            "will not submit for now... your response: {}".format(resp), fg="yellow"
        )
    return shall


def custom_decider(decide_func, idbased, unroll_data=True):
    # we yield until we receive some data via send()
    def decider():
        data = yield
        while True:
            data = (
                yield decide_func(*data, idbased=idbased)
                if unroll_data
                else decide_func(data, idbased=idbased)
            )

    return decider


def interactive_deciders(idbased=False):
    """
    returns a tuple (extend,submit) of already-primed deciders for both
    extension and submission
    """
    extend_decider = custom_decider(decide_rule, idbased)()
    advance_coroutine(extend_decider)  # prime decider

    submit_decider = custom_decider(decide_step, idbased)()
    advance_coroutine(submit_decider)  # prime decider

    return extend_decider, submit_decider



import os
import importlib
import logging
from .localposix import LocalFSProvider

log = logging.getLogger(__name__)

from ..handlers.utils import handler_decorator

provider_deserialize_handlers, provider_deserializer = handler_decorator()


@provider_deserializer("localfs_provider")
def localfs_provider_deserializer(jsondata, deserialization_opts):
    return LocalFSProvider.fromJSON(jsondata, deserialization_opts)


@provider_deserializer("frompython_provider")
def frompython_provider(jsondata, deserialization_opts):
    providerstring = deserialization_opts.get("state_provider", "")
    _, module, providerclass = providerstring.split(":")
    module = importlib.import_module(module)
    providerclass = getattr(module, providerclass)
    provideropts = {}
    return providerclass.fromJSON(jsondata, **provideropts)


@provider_deserializer("fromenv_provider")
def fromenv_provider_deserializer(jsondata, deserialization_opts):
    module = importlib.import_module(os.environ["YADAGE_STATEPROVIDER"])
    return module.load_provider(jsondata)


def load_provider(jsondata, deserialization_opts=None):
    log.debug("load_provider opts %s", deserialization_opts)
    deserialization_opts = deserialization_opts or {}
    if jsondata == None:
        return None
    if "state_provider" in deserialization_opts:
        providerstring = deserialization_opts.get("state_provider", "")
        if providerstring.startswith("py:"):
            return provider_deserialize_handlers["frompython_provider"](
                jsondata, deserialization_opts
            )
    if "YADAGE_STATEPROVIDER" in os.environ:
        return provider_deserialize_handlers["fromenv_provider"](
            jsondata, deserialization_opts
        )

    return provider_deserialize_handlers[jsondata["state_provider_type"]](
        jsondata, deserialization_opts
    )

    raise TypeError("unknown provider type {}".format(jsondata["state_provider_type"]))


providersetup_handlers, providersetup = handler_decorator()


@providersetup("local")
def localfs_provider(dataarg, dataopts):
    import yadage.state_providers.localposix

    return yadage.state_providers.localposix.setup_provider(dataarg, dataopts)


@providersetup("py:")
def formpython_provider(dataarg, dataopts):
    _, module, setupfunc, dataarg = dataarg.split(":", 3)
    module = importlib.import_module(module)
    setupfunc = getattr(module, setupfunc)
    return setupfunc(dataarg, dataopts)


@providersetup("fromenv:")
def fromenv_provider(dataarg, dataopts):
    module = importlib.import_module(os.environ["PACKTIVITY_STATEPROVIDER"])
    return module.setup_provider(dataarg, dataopts)


def state_provider_from_string(dataarg, dataopts=None):
    dataopts = dataopts or {}
    if len(dataarg.split(":", 1)) == 1:
        dataarg = "local:" + dataarg
    for k in providersetup_handlers.keys():
        if dataarg.startswith(k):
            return providersetup_handlers[k](dataarg, dataopts)
    raise RuntimeError("unknown data type %s %s" % (dataarg, dataopts))



import os
from setuptools import setup, find_packages

deps = [
    "adage",
    "packtivity",
    "yadage-schemas",
    "click",
    "psutil",
    "requests[security]>=2.9",
    "pyyaml",
    "jsonref",
    "jsonschema",
    "jsonpointer>=1.10",
    "jsonpath_rw",
    "checksumdir",
    "glob2",
    "jq",
]


if "READTHEDOCS" in os.environ:
    deps = [
        "yadage-schemas",
        "click",
        "psutil",
        "requests[security]>=2.9",
        "pyyaml",
        "jsonref",
        "jsonschema",
        "jsonpointer>=1.10",
        "jsonpath_rw",
        "checksumdir",
        "glob2",
    ]

setup(
    name="yadage",
    version="0.20.1",
    description="yadage - YAML based adage",
    url="",
    author="Lukas Heinrich",
    author_email="lukas.heinrich@cern.ch",
    packages=find_packages(),
    include_package_data=True,
    install_requires=deps,
    extras_require={
        "celery": ["celery", "redis"],
        "viz": [
            # manually adding extras of adage[extra] because of pip
            # issue https://github.com/pypa/pip/issues/3189
            "pydot2",
            "pygraphviz",
            "pydotplus",
        ],
        "develop": [
            "pyflakes",
            "pre-commit",
            'black;python_version>="3.6"',  # Black is Python3 only
            "isort",
            "pytest>=3.2.0",
            "pytest-cov>=2.5.1",
            "python-coveralls",
        ],
    },
    entry_points={
        "console_scripts": [
            "yadage-run=yadage.steering:main",
            "yadage-manual=yadage.manualcli:mancli",
            "yadage-util=yadage.utilcli:utilcli",
        ]
    },
    dependency_links=[],
)
